from .input import Model, Pipeline, ProblemSpec, CV, Custom
from copy import deepcopy
import numpy as np
import pandas as pd
from ..pipeline.BPtPipelineConstructor import get_pipe
from ..default.options.scorers import process_scorers
from .BPtEvaluator import BPtEvaluator
from sklearn.model_selection import check_cv
from .input_operations import Intersection
from pandas.util._decorators import doc
from .CV import inds_from_names
from ..shared_docs import _shared_docs
from .compare import (_compare_check, CompareDict, _merge_compare, Compare)


_base_docs = {}

_base_docs[
    "pipeline"
] = """pipeline : :class:`Pipeline`
        A BPt input class Pipeline to be intialized according
        to the passed dataset and problem_spec.

        This parameter can be either an instance of :class:`Pipeline`,
        :class:`ModelPipeline` or one of the below cases.

        In the case that a single str is passed, it will assumed
        to be a model indicator str and the pipeline used will be:

        ::

            pipeline = Pipeline(Model(pipeline))

        Likewise, if just a Model passed, then the input will be
        cast as:

        ::

            pipeline = Pipeline(pipeline)

    """

_base_docs[
    "dataset"
] = """dataset : :class:`Dataset`
        The Dataset in function should be evaluated in the context of.
        The dataset is as the data source for this operation.

        Arguments within problem_spec can be used to
        select just subsets of data. For example parameter
        scope can be used to select only some columns or
        parameter subjects to select a subset of subjects.
    """

_base_docs["problem_spec"] = _shared_docs['problem_spec']

_base_docs[
    "extra_params"
] = """extra_params : problem_spec or pipeline params, optional
        You may pass as extra arguments to this function any pipeline
        or problem_spec argument as python kwargs style value pairs.

        For example:

        ::

            target=1

        Would override the value of the target parameter
        in the passed problem_spec. Or for example:

        ::

            model=Model('ridge')

    """

_eval_docs = _base_docs.copy()
_eval_docs[
    "cv"
] = """cv : :class:`CV` or :ref:`sklearn CV <cross_validation>`, optional
        This parameter controls what type of cross-validation
        splitting strategy is used. You may pass
        a number of options here.

        - An instance of :class:`CV` representing a
          custom strategy as defined by the BPt style CV.

        - The custom str 'test', which specifies that the
          whole train set should be used to train the pipeline
          and the full test set used to validate it (assuming that
          a train test split has been defined in the underlying dataset)

        - Any valid scikit-learn style option:
          Which include an int to specify the number of folds in a
          (Stratified) KFold, a sklearn CV splitter or an iterable
          yielding (train, test) splits as arrays of indices.

        ::

            default = 5
"""


def pipeline_check(pipeline, error_if_compare=True, **extra_params):

    # Make deep copy
    pipe = deepcopy(pipeline)

    # First check here for compare
    if isinstance(pipe, Compare):
        if error_if_compare:
            raise RuntimeError("This function can't accept Compare arguments!")

        new_pipeline = CompareDict()
        for option in pipe.options:
            option.key = 'pipeline'

            # Get the checked version of the pipeline
            # passing along all extra args to each
            checked_pipe = _base_pipeline_check(option.value, **extra_params)

            # Save this option in the dict
            new_pipeline[option] = checked_pipe

        # Return as compare dict
        return new_pipeline

    # Otherwise, base check
    return _base_pipeline_check(pipe, **extra_params)


def _base_pipeline_check(pipe, **extra_params):

    # If passed pipeline is not Pipeline or ModelPipeline
    # then make input as Pipeline around model
    if not isinstance(pipe, Pipeline):

        # Check for if model str first
        if isinstance(pipe, str):
            pipe = Model(obj=pipe)

        # In case of passed valid single model, wrap in Pipeline
        if isinstance(pipe, Model):
            pipe = Pipeline(steps=[pipe])
        else:
            raise RuntimeError('pipeline must be a Pipeline',
                               ' model str or Model-like')

    # Set any overlapping extra params
    possible_params = pipe._get_param_names()
    valid_params = {key: extra_params[key] for key in extra_params
                    if key in possible_params}
    pipe.set_params(**valid_params)

    # Internal class checks
    pipe._proc_checks()

    return pipe


def _problem_spec_target_check(ps, dataset):

    # Get target col from dataset, this also
    # makes sure scopes and roles are updated.
    targets = dataset.get_cols('target')

    # Check for no targets
    if len(targets) == 0:
        raise IndexError('The passed Dataset is not valid. '
                         'It must have atleast one target defined.')

    # Update target, if passed as int, treat as index
    if isinstance(ps.target, int):
        try:
            ps.target = targets[ps.target]
        except IndexError:
            raise IndexError('target index: ' + repr(ps.target) +
                             ' is out of range, only ' + repr(len(targets)) +
                             ' targets are defined.')

    # If not int, then raise error if invalid or keep as is
    if ps.target not in targets:
        raise IndexError('Passed target: ' + repr(ps.target) + ' does ' +
                         'not have role target or is not loaded.')


def problem_spec_check(problem_spec, dataset, error_if_compare=True,
                       **extra_params):

    # If problem spec is already a CompareDict.
    # Return as is
    if isinstance(problem_spec, CompareDict):
        return problem_spec

    # Set ps to copy of problem spec
    ps = deepcopy(problem_spec)

    # Check if problem_spec is left as default
    if ps == 'default':
        ps = ProblemSpec()

    # Check for any override params
    possible_params = ProblemSpec._get_param_names()
    valid_params = {key: extra_params[key] for key in extra_params
                    if key in possible_params}

    # If any override params - reset checked
    if len(valid_params) > 0:
        ps._checked = False

    # If attr checked, then means the passed
    # problem_spec has already been checked and is already
    # a proc'ed and ready copy.
    if hasattr(ps, '_checked') and getattr(ps, '_checked'):
        return ps

    # Set any overlap params
    ps.set_params(**valid_params)

    # Check for any Compare
    ps = _compare_check(ps)

    # If ps is now a dict, it means there was atleast one Compare
    if isinstance(ps, CompareDict):

        if error_if_compare:
            raise RuntimeError("This function can't accept Compare arguments!")

        return CompareDict({key: _base_ps_check(ps[key], dataset)
                            for key in ps})

    # Otherwise perform base check as usual
    return _base_ps_check(ps, dataset)


def _base_ps_check(ps, dataset):

    # Proc params
    ps._proc_checks()

    # Get target col from dataset, sets target
    # in ps in place.
    _problem_spec_target_check(ps, dataset)

    # Replace problem_spec type w/ correct if passed short hands
    pt = ps.problem_type

    if pt == 'default':
        pt = dataset._get_problem_type(ps.target)
    elif pt == 'b':
        pt = 'binary'
    elif pt == 'c' or pt == 'multiclass':
        pt = 'categorical'
    elif pt == 'f' or pt == 'float':
        pt = 'regression'

    ps.problem_type = pt

    # Convert to scorer obj
    ps.scorer =\
        process_scorers(ps.scorer,
                        problem_type=ps.problem_type)

    # Set checked flag in ps
    setattr(ps, '_checked', True)

    return ps


@doc(**_base_docs)
def get_estimator(pipeline, dataset,
                  problem_spec='default',
                  **extra_params):
    '''Get a sklearn compatible :ref:`estimator<develop>` from a
    :class:`Pipeline`, :class:`Dataset` and :class:`ProblemSpec`.

    Parameters
    -----------
    {pipeline}

    {dataset}

    {problem_spec}

    {extra_params}

    Returns
    --------
    estimator : :ref:`sklearn Estimator <develop>`
        The returned object is a sklearn-compatible estimator.
        It will be either of type BPtPipeline or a BPtPipeline
        wrapped in a search CV object.

    Examples
    ----------
    This example shows how this function can be
    used with Dataset method
    :func:`get_Xy <Dataset.get_Xy>`.

    First we will setup a dataset and a pipeline.

    .. ipython:: python

        import BPt as bp

        data = bp.Dataset([[1, 2, 3], [2, 3, 4], [5, 6, 7]],
                           columns=['1', '2', '3'],
                           targets='3')
        data

        pipeline = bp.Pipeline([bp.Scaler('standard'), bp.Model('linear')])
        pipeline

    Next we can use get_estimator and
    also convert the dataset into a traditional sklearn-style
    X, y input. Note that we are using just the default values
    for problem spec.

    .. ipython:: python

        X, y = data.get_Xy()
        X.shape, y.shape

        estimator = bp.get_estimator(pipeline, data)

    Now we can use this estimator as we would any other sklearn
    style estimator.

    .. ipython:: python

        estimator.fit(X, y)
        estimator.score(X, y)

    '''

    # Use initial prep
    estimator_ps = _initial_prep(pipeline, dataset, problem_spec,
                                 error_if_compare=(True, False),
                                 **extra_params)

    # Reduce estimator_ps to just estimator related
    if isinstance(estimator_ps, CompareDict):

        # Return compare dict w/ only estimator, not estimator_ps tuple
        return CompareDict({key: estimator_ps[key][0] for key in estimator_ps})

    # Otherwise, can just return first element of the tuple
    return estimator_ps[0]


def get_pipeline(pipeline, problem_spec, dataset, error_if_compare=True):

    # Run compare check
    pipeline = _compare_check(pipeline)
    if isinstance(pipeline, CompareDict):
        if error_if_compare:
            raise RuntimeError("This function can't accept Compare arguments!")

        return CompareDict({key: _get_pipeline(
            pipeline[key], problem_spec, dataset)[0]
             for key in pipeline})

    # Otherwise, base
    return _get_pipeline(pipeline, problem_spec, dataset)[0]


def _get_pipeline(pipeline, problem_spec, dataset,
                  has_search=False):
    '''Both pipeline and problem_spec should not be CompareDicts.'''

    # If has search is False, means this is the top level
    # or the top level didn't have a search
    if not has_search:
        has_search = True
        if pipeline.param_search is None:
            has_search = False

    # If either this set of pipeline params or the parent
    # params had search params, then a copy of problem spec with n_jobs set
    # to 1 should be passed to children get pipelines,
    nested_ps = problem_spec
    if has_search:
        nested_ps = deepcopy(problem_spec)
        nested_ps.n_jobs = 1

    # Define a nested check, that iterates through searching for
    # nested model pipeline's
    def nested_check(obj):

        # Case where pipeline is nested as an obj
        if hasattr(obj, 'obj') and isinstance(obj.obj, Pipeline):

            nested_pipe, nested_pipe_params =\
                _get_pipeline(pipeline=obj.obj,
                              problem_spec=nested_ps,
                              dataset=dataset,
                              has_search=has_search)

            # Set obj as nested pipeline
            setattr(obj, 'obj', nested_pipe)

            # Set obj's params as the nested_pipe_params
            setattr(obj, 'params', nested_pipe_params)

        # Case where pipeline is passed as a step
        elif hasattr(obj, 'steps'):
            for i, step in enumerate(obj.steps):
                if isinstance(step, Pipeline):

                    # Call nested get pipeline
                    nested_pipe, nested_pipe_params =\
                        _get_pipeline(pipeline=step,
                                      problem_spec=nested_ps,
                                      dataset=dataset,
                                      has_search=has_search)

                    # If params to pass along, needs to be Model
                    if len(nested_pipe_params) > 0:
                        new_step = Model(nested_pipe,
                                         params=nested_pipe_params)

                    # Otherwise pass as custom
                    else:
                        new_step = Custom(nested_pipe)

                    # Replace step
                    obj.steps[i] = new_step

        # Recursive cases
        if isinstance(obj, list):
            [nested_check(o) for o in obj]
            return

        elif isinstance(obj, tuple):
            (nested_check(o) for o in obj)
            return

        elif isinstance(obj, dict):
            [nested_check(obj[k]) for k in obj]
            return

        elif hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                try:
                    nested_check(getattr(obj, param))
                except AttributeError:
                    return
            return

    # Run nested check on passed pipeline
    nested_check(pipeline)

    # Preproc model
    pipeline =\
        _preproc_pipeline(pipeline, problem_spec, dataset)

    # Return the pipeline
    return get_pipe(pipeline_params=pipeline,
                    problem_spec=problem_spec,
                    dataset=dataset)


def _preproc_pipeline(pipe, ps, dataset):

    # Check imputers default case
    # Check if any NaN in data
    data_cols = dataset._get_cols(scope='data', limit_to=ps.scope)

    is_na = dataset[data_cols].isna().any().any()
    pipe._check_imputers(is_na)

    # Pre-proc param search
    has_param_search = _preproc_param_search(pipe, ps)

    # If there is a param search set for the Model Pipeline,
    # set n_jobs, the value to pass in the nested check, to 1
    nested_ps = deepcopy(ps)
    if has_param_search:
        nested_ps.n_jobs = 1

    def nested_model_check(obj):

        # Check for Model or Ensemble
        if isinstance(obj, Model):
            _preproc_param_search(obj, nested_ps)

        if isinstance(obj, list):
            [nested_model_check(o) for o in obj]
            return

        elif isinstance(obj, tuple):
            (nested_model_check(o) for o in obj)
            return

        elif isinstance(obj, dict):
            [nested_model_check(obj[k]) for k in obj]

        elif hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                try:
                    nested_model_check(getattr(obj, param))
                except AttributeError:
                    return
            return
        return

    # Run nested check
    nested_model_check(pipe)

    # Run nested check for any CV input param objects
    nested_cv_check(pipe, dataset)

    return pipe


def nested_cv_check(obj, dataset):

    def _nested_cv_check(obj):

        # If has cv object, try applying dataset
        # can be either in object or in dict

        # If Object, check if any of the parameters are CV
        if hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                val = getattr(obj, param)

                # If CV proc
                if isinstance(val, CV):
                    setattr(obj, param, val._apply_dataset(dataset))

                # If not, nested check
                else:
                    _nested_cv_check(val)

        # If Dict, check all valyes
        elif isinstance(obj, dict):
            for k in obj:
                val = obj[k]

                # If CV proc
                if isinstance(val, CV):
                    obj[k] = val._apply_dataset(dataset)

                # If not, nested check
                else:
                    _nested_cv_check(val)

        elif isinstance(obj, (list, tuple, set)):
            [_nested_cv_check(o) for o in obj]

    # Run nested check for any CV input param objects
    _nested_cv_check(obj)


def _preproc_param_search(object, ps):

    # Get param search
    param_search = getattr(object, 'param_search')

    # If None, return False and do nothing
    if param_search is None:
        return False

    # If already processed, do nothing, but return True
    if isinstance(param_search, dict):
        return True

    # If a param search, apply ps and dataset + convert to dict
    as_dict = param_search._as_dict(ps=ps)

    # Set new proc'ed
    setattr(object, 'param_search', as_dict)

    return True


def _initial_prep(pipeline, dataset, problem_spec,
                  error_if_compare=True, **extra_params):

    # error if compare can be bool or tuple of bool's
    if isinstance(error_if_compare, tuple):
        eic_ps, eic_pipe = error_if_compare
    else:
        eic_ps, eic_pipe = error_if_compare, error_if_compare

    # Get proc'ed problem spec, w/ possibility that it is
    # returned as a CompareDict.
    ps = problem_spec_check(problem_spec=problem_spec, dataset=dataset,
                            error_if_compare=eic_ps, **extra_params)

    # Want to get pipe, with possibility that returned pipe is CompareDict
    pipe = pipeline_check(pipeline, error_if_compare=eic_pipe,
                          **extra_params)

    # Get the actual pipeline as an estimator
    # where both pipe and ps can be CompareDicts and
    # CompareDicts can still live inside the pipe params

    # Handle if ps if CompareDict and / or pipe is CompareDict
    merged_compare = _merge_compare(pipe=pipe, ps=ps)

    # Base case first
    if not isinstance(merged_compare, CompareDict):

        # Unpack
        c_pipe, c_ps = merged_compare

        # Get the estimator
        estimator = get_pipeline(c_pipe, c_ps,
                                 dataset, error_if_compare=eic_pipe)

        # If estimator is CompareDict, need to add in problem_spec
        # So compare dict stores tuples of estimator problem spec.
        if isinstance(estimator, CompareDict):
            return CompareDict({key: (estimator[key], deepcopy(c_ps))
                               for key in estimator})

        # Otherwise, return as is
        return (estimator, c_ps)

    # CompareDict case
    estimator_ps = CompareDict()
    for key in merged_compare:

        # Get this sub-model
        c_pipe, c_ps = merged_compare[key]
        estimator = get_pipeline(c_pipe, c_ps, dataset,
                                 error_if_compare=eic_pipe)

        # est is either Pipeline or CompareDict
        # If compare dict, add the combination.
        if isinstance(estimator, CompareDict):
            for e_key in estimator:
                estimator_ps[[key, e_key]] =\
                    (deepcopy(estimator[e_key]), deepcopy(c_ps))

        # Otherwise, add as is.
        else:
            estimator_ps[key] = (estimator, c_ps)

    # Return the compare dict
    return estimator_ps


def _sk_prep(pipeline, dataset, problem_spec='default',
             cv=5, **extra_params):

    estimator, ps = _initial_prep(pipeline, dataset, problem_spec,
                                  error_if_compare=True,
                                  **extra_params)

    return _eval_prep(estimator, ps, dataset, cv)


def _eval_prep(estimator, ps, dataset, cv=5):
    '''Internal helper function.'''

    # Check for subjects == 'default' in problem_spec
    if ps.subjects == 'default':

        # If a test set is defined
        if dataset.test_subjects is not None:
            if cv == 'test':
                ps.subjects = 'all'
            else:
                ps.subjects = 'train'

        # Otherwise, subjects = 'all' if no test set defined
        else:
            ps.subjects = 'all'

    # Get X and y - ps should already be init'ed
    X, y = dataset.get_Xy(problem_spec=ps)

    # Save if has n_repeats
    n_repeats = 1
    if hasattr(cv, 'n_repeats'):
        n_repeats = n_repeats

    # If Passed CV class, then need to convert to sklearn compat
    # otherwise, assume that it is sklearn compat and pass as is.
    # @TODO maybe convert to BPtCV then back?
    # to add random state and what not ?
    sk_cv = cv
    if isinstance(cv, CV):

        # Convert from CV class to BPtCV by applying dataset
        bpt_cv = cv._apply_dataset(dataset)

        # Save n_repeats
        n_repeats = bpt_cv.get_n_repeats()

        # Convert from BPtCV to sklearn compat, i.e., just raw index
        sk_cv = bpt_cv.get_cv(X.index,
                              random_state=ps.random_state,
                              return_index=True)

    # Check for passed test case
    elif cv == 'test' or cv == 'Test':

        # Get train and test subjs as intersection of train and ps
        train = dataset.get_subjects(Intersection(['train', ps.subjects]),
                                     return_as='flat index')
        test = dataset.get_subjects(Intersection(['test', ps.subjects]),
                                    return_as='flat index')

        # Set as single iterable.
        sk_cv = [(train, test)]

        # Set as actual index
        sk_cv = inds_from_names(dataset.index, sk_cv)

    # Cast explicitly to sklearn style cv from either user
    # passed input or inds
    is_classifier = ps.problem_type != 'regression'

    # Pass y along only if no NaN's in y.
    sk_cv = check_cv(cv=sk_cv, y=None if pd.isnull(y).any() else y,
                     classifier=is_classifier)

    # Make sure random_state and shuffle are set set, if avaliable attributes.
    if hasattr(sk_cv, 'random_state'):
        setattr(sk_cv, 'random_state', ps.random_state)
    if hasattr(sk_cv, 'shuffle'):
        setattr(sk_cv, 'shuffle', True)

    # Store n_repeats in sk_cv
    setattr(sk_cv, 'n_repeats', n_repeats)

    return estimator, X, y, ps, sk_cv


def _sk_check_y(y):

    if pd.isnull(y).any():
        raise RuntimeError('NaNs are not supported with '
                           'the sklearn compatable evaluate functions. '
                           'See function evaluate instead, which supports '
                           'missing target values, or drop subjects with '
                           'missing target values from the dataset.')


@doc(**_eval_docs)
def cross_val_score(pipeline, dataset,
                    problem_spec='default',
                    cv=5, sk_n_jobs=1, verbose=0,
                    fit_params=None,
                    error_score=np.nan,
                    **extra_params):
    '''This function is a BPt compatible wrapper around
    :func:`sklearn.model_selection.cross_val_score`

    Parameters
    ----------
    {pipeline}

    {dataset}

    {problem_spec}

    {cv}

    sk_n_jobs : int, optional
        The number of jobs as passed to the base sklearn
        cross_val_score. Typically this value
        should be kept at 0, and n_jobs as defined
        through the passed problem_spec used
        to define the number of jobs.

        For added flexibility though, this
        parameter can be used either with
        the n_jobs parameter in problem_spec
        or instead of.

        ::

            default = 1

    verbose : int, optional
        The verbosity level as passed to the sklearn function.

        ::

            default = 0

    fit_params : dict, optional
        Parameters to pass to the fit method of the estimator.

        ::

            default = None


    error_score : 'raise' or numeric, optional
        Base sklearn func parameter.

        ::

            default = np.nan

    {extra_params}

    See Also
    ---------
    cross_validate : To run on multiple metrics and other options.
    evaluate : The BPt style similar function with extra options.

    '''

    # Get sk compat pieces
    estimator, X, y, ps, sk_cv =\
        _sk_prep(pipeline=pipeline, dataset=dataset,
                 problem_spec=problem_spec, cv=cv, **extra_params)

    # Make sure no NaN's in y
    _sk_check_y(y)

    # Just take first scorer if dict
    if isinstance(ps.scorer, dict):
        ps.scorer = ps.scorer[list(ps.scorer)[0]]

    from sklearn.model_selection import cross_val_score
    return cross_val_score(estimator=estimator, X=X, y=y, scoring=ps.scorer,
                           cv=sk_cv, n_jobs=sk_n_jobs, verbose=verbose,
                           fit_params=fit_params, error_score=error_score)


@doc(**_eval_docs)
def cross_validate(pipeline, dataset,
                   problem_spec='default',
                   cv=5, sk_n_jobs=1, verbose=0,
                   fit_params=None,
                   return_train_score=False,
                   return_estimator=False,
                   error_score=np.nan,
                   **extra_params):
    '''This function is a BPt compatible wrapper around
    :func:`sklearn.model_selection.cross_validate`

    Parameters
    ----------
    {pipeline}

    {dataset}

    {problem_spec}

    {cv}

    sk_n_jobs : int, optional
        The number of jobs as passed to the base sklearn
        cross_val_score. Typically this value
        should be kept at 0, and n_jobs as defined
        through the passed problem_spec used
        to define the number of jobs.

        For added flexibility though, this
        parameter can be used either with
        the n_jobs parameter in problem_spec
        or instead of.

        ::

            default = 1

    verbose : int, optional
        The verbosity level as passed to the sklearn function.

        ::

            default = 0

    fit_params : dict, optional
        Parameters to pass to the fit method of the estimator.

        ::

            default = None

    return_train_score : bool, optional
        Whether to include train scores.

        ::

            default = False

    return_estimator : bool, optional
        Whether to return the estimators fitted on each split.

        ::

            default = False


    error_score : 'raise' or numeric, optional
        Base sklearn func parameter.

        ::

            default = np.nan

    {extra_params}

    See Also
    ---------
    cross_val_score : Simplified version of this function.
    evaluate : The BPt style similar function with extra options.

    '''

    # Get sk compat pieces
    estimator, X, y, ps, sk_cv =\
        _sk_prep(pipeline=pipeline, dataset=dataset,
                 problem_spec=problem_spec, cv=cv, **extra_params)

    # Make sure no NaN's in y.
    _sk_check_y(y)

    from sklearn.model_selection import cross_validate
    return cross_validate(estimator=estimator, X=X, y=y, scoring=ps.scorer,
                          cv=sk_cv, n_jobs=sk_n_jobs, verbose=verbose,
                          fit_params=fit_params,
                          return_train_score=return_train_score,
                          return_estimator=return_estimator,
                          error_score=error_score)


@doc(**_eval_docs)
def evaluate(pipeline, dataset,
             problem_spec='default',
             cv=5,
             progress_bar=True,
             store_preds=True,
             store_estimators=True,
             store_timing=True,
             decode_feat_names=True,
             eval_verbose=0,
             progress_loc=None,
             **extra_params):
    ''' This method is used to evaluate a model pipeline
    with cross-validation.

    Parameters
    -----------
    {pipeline}

    {dataset}

    {problem_spec}

    {cv}

    progress_bar : bool, optional
        If True, then an progress
        bar will be displaying showing fit
        progress across
        evaluation splits.

        ::

            default = True

    store_preds : bool, optional
        If set to True, the returned :class:`BPtEvaluator` will store
        the saved predictions under :data:`BPt.BPtEvaluator.preds`.
        This includes a saved copy of the true target values as well.

        If False, the `preds` parameter will be empty and it
        will not be possible to use some related functions.

        ::

            default = True

    store_estimators : bool, optional
        If True, then the returned :class:`BPtEvaluator`
        will store the fitted estimators from evaluation
        under :data:`BPt.BPtEvaluator.estimators`.

        If False, the `estimators` parameter will be empty,
        and it will not be possible to access measures of
        feature importance as well.


        ::

            default = True

    store_timing : bool, optional
        If True, then the returned :class:`BPtEvaluator`
        will store the time it took to fit and score the pipeline
        under :data:`BPt.BPtEvaluator.timing`.

        ::

            default = True

    decode_feat_names : bool, optional
        If True, then the :data:`BPt.BPtEvaluator.feat_names`
        as computed during evaluation will try to use the original
        values as first loaded to inform their naming. Note that
        this is only relevant assuming that :class:`Dataset` was
        used to encode one or more columns in the first place.

        ::

            default = True

    eval_verbose : int, optional
        The requested verbosity of the evaluator.
        0 or greater means just warnings, 1 or greater
        for some info and 2 and greater for even more.

        Set to negative values to mute warnings.
        Specifically, setting to -1 will mute all
        warnings as generated by the call to evaluate,
        and even further, you can set to -2 or lower, which
        will mute all warnings regardless of where they are generated from.

        Note: This parameter is called eval_verbose, as the
        pipeline has an argument called `verbose`, which can be used
        to set verbosity for the pipeline.

        ::

            default = 0

    progress_loc : str or None, optional
        This parameter is not currently implemented.

        ::

            default = None

    {extra_params}

    Returns
    ---------
    evaluator : :class:`BPtEvaluator`
        Returns an instance of the :class:`BPtEvaluator`
        class. This object stores a wealth of information,
        including the scores from this evaluation as well
        as other utilities including functions for calculating
        feature importances from trained models.

    See Also
    ---------
    cross_val_score : Similar sklearn style function.
    cross_validate : Similar sklearn style function.
    Compare : Input class for specifying comparisons between
        parameter options.

    Notes
    ------
    | This function can accept within the `pipeline` and `problem_spec`
        parameters the special input :class:`Compare` class. This
        option is designed for explicitly running evaluate multiple
        times under different configurations.

    | This function supports predictions on an underlying
        target with missing values. It does this by automatically
        moving any data points with a NaN in the target to the validation
        set (keeping in mind this is done after the fold is computed by CV,
        so final size may vary). While subjects with missing values will
        obviously not contribute to the validation score,
        as long as `store_preds`
        is set to True, predictions will still be made for these subjects.
    '''

    params = {'cv': cv,
              'decode_feat_names': decode_feat_names,
              'progress_bar': progress_bar,
              'store_preds': store_preds,
              'store_estimators': store_estimators,
              'store_timing': store_timing,
              'eval_verbose': eval_verbose,
              'progress_loc': progress_loc,
              }

    # Get the estimator and problem spec,
    # w/ option for returned
    # value as a CompareDict
    estimator_ps =\
        _initial_prep(pipeline, dataset, problem_spec,
                      error_if_compare=False, **extra_params)

    # Base case
    if not isinstance(estimator_ps, CompareDict):
        estimator, ps = estimator_ps
        return _evaluate(estimator=estimator, dataset=dataset, ps=ps, **params)

    # Set at start to number of runs
    if progress_bar:
        compare_bars = len(estimator_ps)
    else:
        compare_bars = None

    # Compare dict case
    evaluators = CompareDict()
    for key in estimator_ps:

        if eval_verbose >= 1:
            print('Running Compare:', key, flush=True)

        # Unpack
        c_estimator, c_ps = estimator_ps[key]

        # Evaluate this option
        evaluator = _evaluate(estimator=c_estimator,
                              dataset=dataset, ps=c_ps,
                              compare_bars=compare_bars,
                              **params)

        # Update compare bars to be the the compare bars
        # set by last run
        compare_bars = evaluator.compare_bars

        # Then reset stored attribute,
        # so the evaluator object can be pickled
        evaluator.compare_bars = None

        # Add to compare dict
        evaluators[key] = evaluator

    # Close compare bars
    if compare_bars is not None:
        for bar in compare_bars:
            bar.n = bar.total
            bar.refresh()
            bar.close()

    return evaluators


def _evaluate(estimator, dataset, ps, cv,
              decode_feat_names, compare_bars=None,
              **verbose_args):

    estimator, X, y, ps, sk_cv =\
        _eval_prep(estimator, ps, dataset, cv)

    # Check decode feat_names arg, if True, pass along encoders
    encoders = None
    if decode_feat_names:
        encoders = dataset.encoders

    # Init evaluator
    evaluator = BPtEvaluator(estimator=estimator, ps=ps,
                             encoders=encoders,
                             compare_bars=compare_bars,
                             **verbose_args)

    # Call eval on the evaluator
    evaluator._eval(X, y, sk_cv)

    # Return the BPtEvaluator object
    return evaluator



from copy import deepcopy
import numpy as np
import inspect
from ..util import is_array_like


def args_repr(args, kwargs):

    reprs = [repr(arg) for arg in args]
    sorted_keys = sorted(list(kwargs))
    reprs += [key + '=' + repr(kwargs[key]) for key in sorted_keys]
    return ', '.join(reprs)


def get_obj_and_params(obj_str, OBJS, extra_params, params):

    from .params.default_params import get_base_params

    # First get the object, and process the base params!
    try:
        obj, param_names = OBJS[obj_str]
    except KeyError:
        raise KeyError(repr(obj_str) + ' does not exist!')

    # If params is a str, change it to the relevant index
    if isinstance(params, str):
        try:
            params = param_names.index(params)
        except ValueError:
            print('str', params, 'passed, but not found as an option for',
                  obj_str)
            print('Setting to default base params setting instead!')
            params = 0

    # If passed param ind is a dict, assume that user passed
    if isinstance(params, dict):
        base_params = params.copy()

    # If not a dict passed, grab the param name, then params
    else:

        # Get the actual params
        try:
            param_name = param_names[params]
        except IndexError:
            print('Invalid param ind', params, 'passed for', obj_str)
            print('There are only', len(param_names), 'valid param options.')
            print('Setting to default base params setting instead!')
            param_name = param_names[0]

        base_params = get_base_params(param_name)

    # Process rest of params by search type, and w.r.t to extra params
    non_search_params, params =\
        process_params_by_type(obj_str, base_params, extra_params)

    return obj, non_search_params, params


def process_params_by_type(obj_str, base_params, extra_params):
    '''base params is either a dict or 0'''

    from .params.Params import Params
    from .params.default_params import proc_params

    non_search_params, params = {}, {}

    # Start with params as copy of passed base_params
    params = deepcopy(base_params)

    # If params is passed as 0, treat as no search params
    if params == 0:
        params = {}

    # Extract any non-search params from params
    param_keys = list(params)
    for p in param_keys:

        # If not a special search param
        if not isinstance(params[p], Params):
            non_search_params[p] = params.pop(p)

    # Append obj_str to all params still in params
    # as these are the search params
    params = proc_params(params, prepend=obj_str)

    # Process extra params
    non_search_params, params =\
        proc_extra_params(extra_params=extra_params,
                          non_search_params=non_search_params,
                          params=params)

    return non_search_params, params


def proc_extra_params(extra_params, non_search_params, params=None):

    if extra_params is None or len(extra_params) == 0:
        return non_search_params, params

    for key in extra_params:
        non_search_params[key] = deepcopy(extra_params[key])

        # Override value w/ extra params if also in params
        if params is not None and key in params:
            del params[key]

    return non_search_params, params


def proc_type_dep_str(in_strs, avaliable, problem_type):
    '''Helper function to perform str correction on
    underlying problem type dependent input, e.g., for
    ensemble_types, and to update extra params
    and check to make sure input is valid ect...'''

    as_arr = True
    if not is_array_like(in_strs):
        as_arr = False
        in_strs = [in_strs]

    if not check_avaliable(in_strs, avaliable, problem_type):
        in_strs = proc_input(in_strs)

        if not check_avaliable(in_strs, avaliable, problem_type):
            raise RuntimeError(in_strs, 'are not avaliable for '
                               'this problem type.'
                               'This may be due to the requested object '
                               'being an optional dependency! Check to make '
                               'sure you have the relevant python library '
                               'installed '
                               'and that the passed str contains no typos!')

    avaliable_by_type = get_a_by_type(avaliable, in_strs, problem_type)
    final_strs = [avaliable_by_type[in_str] for in_str in in_strs]

    if as_arr:
        return final_strs
    return final_strs[0]


def get_a_by_type(avaliable, in_strs, problem_type):

    avaliable_by_type = avaliable[problem_type]

    for s in in_strs:
        if 'Custom ' in s:
            avaliable_by_type[s] = s

    return avaliable_by_type


def check_avaliable(in_strs, avaliable, problem_type):

    avaliable_by_type = get_a_by_type(avaliable, in_strs, problem_type)

    check = np.array([m in avaliable_by_type for
                      m in in_strs]).all()

    return check


def proc_input(in_vals):
    '''Performs common preproc on a list of str's or
    a single str.'''

    if isinstance(in_vals, list):
        for i in range(len(in_vals)):
            in_vals[i] = proc_str_input(in_vals[i])
    else:
        in_vals = proc_str_input(in_vals)

    return in_vals


def proc_str_input(in_str):

    if not isinstance(in_str, str):
        return in_str

    # Make sure lower-case
    in_str = in_str.lower()

    # Remove regressor or classifier
    chunk_replace_dict = {' regressor': '',
                          ' classifier': '',
                          '_regressor': '',
                          '_classifier': ''}

    for chunk in chunk_replace_dict:
        in_str = in_str.replace(chunk, chunk_replace_dict[chunk])

    return in_str


def get_possible_init_params(model):
    '''Helper function to grab the names of valid arguments to
    classes init

    Parameters
    ----------
    model : object
        The object to inspect

    Returns
    ----------
        All valid parameters to the model
    '''

    try:
        return model._get_param_names()
    except AttributeError:
        pos_params = dict(inspect.getmembers(model.__init__.__code__))
        return pos_params['co_varnames']


def all_from_avaliable(avaliable):

    a = set()
    for pt in avaliable:
        for key in avaliable[pt]:
            a.add(key)

    return a


def all_from_objects(objects):

    a = set()
    for key in objects:
        a.add(key)

    return a


def coarse_any_obj_check(in_str):
    from .options.all_keys import get_all_keys

    all_keys = get_all_keys()

    # Get proc str
    proc_str = proc_str_input(in_str)

    if proc_str not in all_keys:
        raise RuntimeError('Passed obj=' + str(in_str) + ' does not '
                           'correspond '
                           'to any avaliable default options from BPt!')



import os
import tempfile
import numpy as np
from ...dataset.data_file import DataFile
from ...main.input import ProblemSpec
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection._base import SelectorMixin
from ...main.input import ParamSearch
from ...main.CV import BPtCV, CVStrategy
from ...dataset.Dataset import Dataset


class ToFixedTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, to, n_jobs=1):
        self.to = to
        self.n_jobs = n_jobs

    def fit(self, X, y):
        self.n_features_in_ = X.shape[1]
        return self

    def transform(self, X):
        X_trans = np.zeros(X.shape)
        X_trans[:] = self.to
        return X_trans


class IdentityListLoader(BaseEstimator, TransformerMixin):
    needs_mapping = True

    def __init__(self):
        pass

    def fit(self, X, y=None, mapping=None):
        assert mapping is not None
        return self

    def transform(self, X):

        assert isinstance(X, list)

        X_trans = []
        for x in X:
            X_trans.append(x.flatten())

        return np.array(X_trans)


class FakeSelector(SelectorMixin, BaseEstimator):

    def __init__(self, mask):
        self.mask = mask

    def fit(self, X, y):
        return self

    def _get_support_mask(self):
        return self.mask


def get_param_search():

    param_search = ParamSearch(search_type='RandomSearch',
                               cv='default',
                               n_iter=10,
                               scorer='default',
                               weight_scorer=False,
                               mp_context='loky',
                               n_jobs='default',
                               dask_ip=None,
                               memmap_X=False,
                               search_only_params=None,
                               progress_loc=None)

    ps = ProblemSpec(random_state=1,
                     n_jobs=2,
                     problem_type='regression')

    ps_dict = param_search._as_dict(ps)
    ps_dict['cv'] = BPtCV(splits=3, n_repeats=1,
                          cv_strategy=CVStrategy(), splits_vals=None)

    return ps_dict


def get_temp_files(n):

    temp_dr = tempfile.gettempdir()
    return [os.path.join(temp_dr, 'test_' + str(i) + '.npy') for i in range(n)]


def get_fake_mapping(n):

    locs = get_temp_files(n)

    mapping = {}
    for i, loc in enumerate(locs):
        data = np.zeros((2, 2))
        data[:] = i

        np.save(loc, data)

        mapping[i] = DataFile(loc=loc, load_func=np.load)

    return mapping


def clean_fake_mapping(n):

    locs = get_temp_files(n)
    for loc in locs:
        os.unlink(loc)


def get_fake_data_dataset(data_keys=None,
                          cat_keys=None):

    if data_keys is None:
        data_keys = []

    if cat_keys is None:
        cat_keys = []

    dataset = Dataset()

    for key in data_keys:
        dataset[key] = []
        dataset.set_role(key, 'data', inplace=True)

    for key in cat_keys:
        dataset[key] = []
        dataset.set_role(key, 'data', inplace=True)
        dataset.add_scope(key, 'category', inplace=True)

    dataset._check_scopes()

    return dataset




P = {}

# Imputers
P['mean imp'] = {'strategy': 'mean'}
P['median imp'] = {'strategy': 'median'}
P['most freq imp'] = {'strategy': 'most_frequent'}
P['constant imp'] = {'strategy': 'constant'}
P['iterative imp'] = {'initial_strategy': 'mean',
                      'skip_complete': True}



from pandas.core.arrays import ExtensionArray
from .data_file_type import DataFileDtype
from .data_file import DataFile
import numpy as np
from pandas._libs import lib


class DataFileArray(ExtensionArray):

    _typ = "datafilearray"
    _scalar_type = DataFile

    ndim = 1
    can_hold_na = True

    _dtype = DataFileDtype

    def __init__(self, data):

        # If passed DataFileArray
        if isinstance(data, DataFileArray):
            self._set_data(data._data)

        # Otherwise set
        self._set_data(data)

    def _set_data(self, data):

        # Data should be a numpy area with dtype
        # object, filled with DataFile objects

        self._data = np.empty(len(data), dtype=object)
        for i, file in enumerate(data):
            if not isinstance(file, self._scalar_type):
                self._data[i] = self._scalar_type(file)
            else:
                self._data[i] = file

    @classmethod
    def _from_sequence(cls, scalars, *, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of scalars.
        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type`` or be converted into this type
            in this method.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.
        Returns
        -------
        DataFileArray
        """

        result = cls(data=scalars)

        return result

    @classmethod
    def _from_sequence_of_strings(cls, strings, *, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of strings.
        .. versionadded:: 0.24.0
        Parameters
        ----------
        strings : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.
        Returns
        -------
        ExtensionArray
        """

        return cls()

    def _from_factorized(cls, values, original):
        return cls()

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        raise NotImplementedError(
            f"{type(self)} does not implement __setitem__.")

    def __len__(self) -> int:
        """
        Length of this array
        Returns
        -------
        length : int
        """
        return len(self._data)

    def __iter__(self):
        """
        Iterate over elements of the array.
        """
        # This needs to be implemented so that pandas recognizes extension
        # arrays as list-like. The default implementation makes successive
        # calls to ``__getitem__``, which may be slower than necessary.
        for i in range(len(self)):
            yield self[i]

    def __eq__(self, other):
        """
        Return for `self == other` (element-wise equality).
        """

        return self._data == other._data

    def __ne__(self, other):
        """
        Return for `self != other` (element-wise in-equality).
        """
        return ~(self._data == other._data)

    def to_numpy(self, dtype=None, copy=False, na_value=lib.no_default):
        print('to numpy !!!')

    # ------------------------------------------------------------------------
    # Required attributes
    # ------------------------------------------------------------------------

    @property
    def shape(self):
        """
        Return a tuple of the array dimensions.
        """
        return self._data.shape

    @property
    def size(self) -> int:
        """
        The number of elements in the array.
        """
        return np.prod(self.shape)

    @property
    def ndim(self) -> int:
        """
        Extension Arrays are only allowed to be 1-dimensional.
        """
        return 1

    @property
    def nbytes(self) -> int:
        return self._data.nbytes

    @property
    def dtype(self):
        return DataFileDtype

    def astype(self, dtype, copy=True):
        """
        Cast to a NumPy array with 'dtype'.
        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, default True
            Whether to copy the data, even if not necessary. If False,
            a copy is made only if the old dtype does not match the
            new dtype.
        Returns
        -------
        array : ndarray
            NumPy ndarray with 'dtype' for its dtype.
        """
        print('astype!')

    def copy(self):
        """
        Return a copy of the array.
        Returns
        -------
        ExtensionArray
        """

        return DataFileArray(self._data.copy())



from sklearn.base import BaseEstimator
from .base import _get_est_fit_params, _needs, _get_est_trans_params
from sklearn.model_selection import GridSearchCV
from sklearn.utils.metaestimators import if_delegate_has_method
import warnings
from os.path import dirname, abspath, exists

import numpy as np
from numpy.random import RandomState
import nevergrad as ng

from concurrent import futures
import multiprocessing as mp

from sklearn.base import clone
from copy import deepcopy
import os
import pandas as pd

from .helpers import to_memmap, from_memmap, get_grid_params
from loky import get_reusable_executor


class BPtSearchCV(BaseEstimator):

    _needs_mapping = True
    _needs_fit_index = True
    name = 'search'

    def __init__(self, estimator=None, ps=None,
                 param_distributions=None,
                 n_jobs=1,
                 random_state=None):

        self.estimator = estimator
        self.ps = ps
        self.param_distributions = param_distributions
        self.n_jobs = n_jobs
        self.random_state = random_state

    @property
    def _needs_transform_index(self):
        return _needs(self.estimator, '_needs_transform_index',
                      'transform_index', 'transform')

    @property
    def n_jobs(self):
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs):

        # Store in self._n_jobs
        self._n_jobs = n_jobs

        # If passed n_jobs, then propegate
        # n_jobs == 1 to base models.
        if hasattr(self.estimator, 'n_jobs'):
            setattr(self.estimator, 'n_jobs', 1)

    @property
    def random_state(self):
        return self._random_state

    @random_state.setter
    def random_state(self, random_state):

        # Store
        self._random_state = random_state

        # Propegate
        if hasattr(self.estimator, 'random_state'):
            setattr(self.estimator, 'random_state', random_state)

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.
        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            try:
                value = getattr(self, key)
            except AttributeError:
                warnings.warn('From version 0.24, get_params will raise an '
                              'AttributeError if a parameter cannot be '
                              'retrieved as an instance attribute. Previously '
                              'it would return None.',
                              FutureWarning)
                value = None
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    @property
    def feature_importances_(self):
        if hasattr(self.best_estimator_, 'feature_importances_'):
            return getattr(self.best_estimator_, 'feature_importances_')
        return None

    @property
    def coef_(self):
        if hasattr(self.best_estimator_, 'coef_'):
            return getattr(self.best_estimator_, 'coef_')
        return None

    @property
    def _final_estimator(self):
        if hasattr(self.best_estimator_, '_final_estimator'):
            return self.best_estimator_._final_estimator
        return None

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict(self, X):
        return self.best_estimator_.predict(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_log_proba(self, X):
        return self.best_estimator_.predict_log_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_proba(self, X):
        return self.best_estimator_.predict_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def decision_function(self, X):
        return self.best_estimator_.decision_function(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform(self, X, transform_index=None):

        trans_params = _get_est_trans_params(
            self.best_estimator_,
            transform_index=transform_index)

        return self.best_estimator_.transform(X, **trans_params)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def inverse_transform(self, Xt):
        return self.best_estimator_.inverse_transform(Xt)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def score(self, X, y=None, sample_weight=None):
        return self.best_estimator_.score(X=X, y=y,
                                          sample_weight=sample_weight)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform_df(self, X_df, encoders=None):
        return self.best_estimator_.transform_df(X_df, encoders=encoders)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform_feat_names(self, X_df, encoders=None):
        return self.best_estimator_.transform_feat_names(X_df,
                                                         encoders=encoders)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def inverse_transform_FIs(self, fis):
        return self.best_estimator_.inverse_transform_FIs(fis)

    def _set_cv(self, fit_index):

        # Set cv based on fit_index
        self.cv_subjects, self.cv_inds =\
            self.ps['cv'].get_cv(fit_index,
                                 self.ps['random_state'],
                                 return_index='both')

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        # Conv from dataframe if dataframe.
        if isinstance(X, pd.DataFrame):
            fit_index = X.index
            X = np.array(X)
        if isinstance(y, (pd.DataFrame, pd.Series)):
            y = np.array(y)

        # Make sure train data index is passed
        if fit_index is None:
            raise RuntimeWarning('SearchCV Object must be passed a '
                                 'fit_index! Or passed X as '
                                 'a DataFrame.')

        if self.ps['verbose'] > 1:
            print('Fit Search CV, len(fit_index) == ',
                  len(fit_index),
                  'has mapping == ', 'mapping' in fit_params,
                  'X.shape ==', X.shape)

        # Set the search cv passed on passed fit_index
        self._set_cv(fit_index)

        # Run different fit depending on type of search
        if self.ps['search_type'] == 'grid':
            self.fit_grid(X=X, y=y, mapping=mapping,
                          fit_index=fit_index,
                          **fit_params)
        else:
            self.fit_nevergrad(X=X, y=y, mapping=mapping,
                               fit_index=fit_index,
                               **fit_params)

        return self


class BPtGridSearchCV(BPtSearchCV):

    def fit_grid(self, X, y=None, mapping=None,
                 fit_index=None, **fit_params):

        # Conv nevergrad to grid compat. param grid
        param_grid = get_grid_params(self.param_distributions)

        # Fit GridSearchCV object
        self.search_obj_ = GridSearchCV(estimator=self.estimator,
                                        param_grid=param_grid,
                                        scoring=self.ps['scorer'],
                                        n_jobs=self.n_jobs,
                                        cv=self.cv_inds,
                                        refit=True,
                                        verbose=self.ps['verbose'])

        # Generate the fit params to pass
        f_params = _get_est_fit_params(
            self.estimator,
            mapping=mapping,
            fit_index=fit_index,
            other_params=fit_params)

        # Hack to support scoring for needs_transform
        if _needs(self.estimator, '_needs_transform_index',
                  'transform_index', 'transform'):
            score_X = pd.DataFrame(X, index=fit_index)
            self.search_obj_.fit(score_X, y, **f_params)

        else:
            self.search_obj_.fit(X, y, **f_params)

        return self

    @property
    def n_features_in_(self):
        return self.search_obj_.n_features_in_

    @property
    def classes_(self):
        return self.search_obj_.classes_

    @property
    def best_estimator_(self):
        return self.search_obj_.best_estimator_

    @property
    def best_score_(self):
        return self.search_obj_.best_score_

    @property
    def best_params_(self):
        return self.search_obj_.best_params_


class ProgressLogger():

    def __init__(self, loc):
        self.loc = loc

    def __call__(self, optimizer=None, candidate=None, value=None):

        # If progress loc parent folder is removed, stop program
        if not exists(dirname(abspath(self.loc))):
            raise SystemExit('Folder where progress is stored was removed!')

        with open(self.loc, 'a') as f:
            f.write('params,')


def ng_cv_score(X, y, estimator, scoring, weight_scorer, cv_inds, cv_subjects,
                mapping, fit_params, search_only_params, **search_params):

    # If passing memmap
    if isinstance(X, tuple):
        X = from_memmap(X)

    cv_scores = []
    for i in range(len(cv_inds)):
        tr_inds, test_inds = cv_inds[i]

        # Clone estimator & set search params
        estimator = clone(estimator)
        estimator.set_params(**search_params)

        # For each search only param, try to update, if invalid, just skip
        for key in search_only_params:
            try:
                estimator.set_params(**{key: search_only_params[key]})
            except ValueError:
                pass

        # Adds mapping / train data index if needed
        f_params = _get_est_fit_params(
            estimator,
            mapping=mapping,
            fit_index=cv_subjects[i][0],
            other_params=fit_params)

        # Fit estimator on train
        estimator.fit(X[tr_inds], y[tr_inds], **deepcopy(f_params))

        # Get the score, but scoring return high values as better,
        # so flip sign.

        # Hack to allow passing info on transform_index.
        if _needs(estimator, '_needs_transform_index',
                  'transform_index', 'transform'):
            score_X = pd.DataFrame(X[test_inds], index=cv_subjects[i][1])
            score = -scoring(estimator, score_X, y[test_inds])

        else:
            score = -scoring(estimator, X[test_inds], y[test_inds])

        # score = -scoring(estimator, X[test_inds], y[test_inds])
        cv_scores.append(score)

    if weight_scorer:
        weights = [len(cv_inds[i][1]) for i
                   in range(len(cv_inds))]
        return np.average(cv_scores, weights=weights)
    else:
        return np.mean(cv_scores)


class NevergradSearchCV(BPtSearchCV):

    # @TODO add more info from verbose!
    # @TODO add cv_results_, best_index_
    # other features of sklearn style.

    @property
    def n_features_in_(self):
        return self.best_estimator_.n_features_in_

    @property
    def classes_(self):
        return self.best_estimator_.classes_

    def get_instrumentation(self, X, y, mapping, fit_params, client):

        X_file = None

        if client is None:

            # Check for memmap X, only if no client
            if self.ps['memmap_X']:
                X_mem = to_memmap(X)
            else:
                X_mem = X

            instrumentation =\
                ng.p.Instrumentation(X_mem, y, self.estimator,
                                     self.ps['scorer'],
                                     self.ps['weight_scorer'],
                                     self.cv_inds,
                                     self.cv_subjects, mapping,
                                     fit_params,
                                     self.ps['search_only_params'],
                                     **self.param_distributions)

        # If using dask client, pre-scatter some big memory fixed params
        else:
            X_s = client.scatter(X)
            y_s = client.scatter(y)
            cv_inds_s = client.scatter(self.cv_inds)
            cv_subjects_s = client.scatter(self.cv_subjects)

            instrumentation =\
                ng.p.Instrumentation(X_s, y_s, self.estimator,
                                     self.ps['scorer'],
                                     self.ps['weight_scorer'],
                                     cv_inds_s,
                                     cv_subjects_s, mapping,
                                     fit_params,
                                     self.ps['search_only_params'],
                                     **self.param_distributions)

        return instrumentation, X_file

    def get_optimizer(self, instrumentation):

        try:
            opt = ng.optimizers.registry[self.ps['search_type']]

        # If not found, look for in experimental variants
        except KeyError:
            import nevergrad.optimization.experimentalvariants
            opt = ng.optimizers.registry[self.ps['search_type']]

        optimizer = opt(parametrization=instrumentation,
                        budget=self.ps['n_iter'],
                        num_workers=self.n_jobs)

        # Set random state is defined
        if isinstance(self.random_state, int):
            optimizer.parametrization.random_state =\
                RandomState(self.random_state)

        elif self.random_state is not None:
            optimizer.parametrization.random_state =\
                self.random_state

        if self.ps['progress_loc'] is not None:
            logger = ProgressLogger(self.ps['progress_loc'])
            optimizer.register_callback('tell', logger)

        return optimizer

    def run_search(self, optimizer, client):

        # n_jobs 1, always local
        if self.n_jobs == 1:
            recommendation = optimizer.minimize(ng_cv_score,
                                                batch_mode=False)

        # If generated client
        elif client is not None:
            recommendation = optimizer.minimize(ng_cv_score,
                                                executor=client,
                                                batch_mode=False)

        # Otherwise use futures pool executor
        else:

            if self.ps['mp_context'] == 'loky':

                executor = get_reusable_executor(
                    max_workers=self.n_jobs, timeout=120)

                recommendation = optimizer.minimize(ng_cv_score,
                                                    executor=executor,
                                                    batch_mode=False)

            try:
                with futures.ProcessPoolExecutor(
                  max_workers=self.n_jobs,
                  mp_context=mp.get_context(self.ps['mp_context'])) as ex:

                    recommendation = optimizer.minimize(ng_cv_score,
                                                        executor=ex,
                                                        batch_mode=False)
            except RuntimeError:
                raise(RuntimeError('Try changing the mp_context'))

        # Save best search search score
        # "optimistic", "pessimistic", "average"
        # and best params
        self.best_score_ = optimizer.current_bests["optimistic"].mean
        self.best_params_ = recommendation.kwargs

        return recommendation

    def fit_nevergrad(self, X, y=None, mapping=None,
                      fit_index=None, **fit_params):

        # Check if need to make dask client
        # Criteria is greater than 1 job, and passed as dask_ip of non-None
        if self.n_jobs > 1 and self.ps['dask_ip'] is not None:
            from dask.distributed import Client
            client = Client(self.ps['dask_ip'])
        else:
            client = None

        # Get the instrumentation
        instrumentation, X_file =\
            self.get_instrumentation(X, y, mapping=mapping,
                                     fit_params=fit_params,
                                     client=client)

        # Get the optimizer
        optimizer = self.get_optimizer(instrumentation)

        # Run the search
        recommendation = self.run_search(optimizer, client)

        # If X was mem mapped, unlink here
        if X_file is not None:
            os.unlink(X_file)

        # Fit best est, w/ best params
        self.fit_best_estimator(recommendation, X, y, mapping,
                                fit_index, fit_params)

        return self

    def fit_best_estimator(self, recommendation,  X, y, mapping,
                           fit_index, fit_params):

        # Fit best estimator, w/ found best params
        self.best_estimator_ = clone(self.estimator)
        self.best_estimator_.set_params(**recommendation.kwargs)

        # Add in mapping / train data index to fit params if needed
        f_params = _get_est_fit_params(
            self.best_estimator_,
            mapping=mapping,
            fit_index=fit_index,
            other_params=fit_params)

        # Fit
        self.best_estimator_.fit(X, y, **f_params)


def wrap_param_search(param_search, model_obj, model_params):

    if param_search is None:
        return model_obj, model_params

    name = model_obj[0]
    prepend = name + '__'

    # Remove the relevant model params
    # and put in m_params
    m_params = {}
    model_param_names = list(model_params)
    for param in model_param_names:

        if param.startswith(prepend):
            m_params[param.replace(prepend, '', 1)] =\
                model_params.pop(param)

    # Get search cv
    search_obj = get_search_cv(
        estimator=model_obj[1],
        param_search=param_search,
        param_distributions=m_params)

    # Create the wrapper nevergrad CV model
    return (name + '_SearchCV', search_obj), model_params


def get_search_cv(estimator, param_search,
                  param_distributions):

    # Determine which CV model to make
    if param_search['search_type'] == 'grid':
        SearchCV = BPtGridSearchCV
    else:
        SearchCV = NevergradSearchCV

    search_obj = SearchCV(
        estimator=estimator,
        ps=param_search,
        param_distributions=param_distributions,
        n_jobs=param_search['n_jobs'],
        random_state=param_search['random_state'])

    return search_obj



from sklearn.neural_network import MLPRegressor, MLPClassifier
import numpy as np


class MLPRegressor_Wrapper(MLPRegressor):

    def _fit(self, X, y, incremental=False):

        self.hidden_layer_sizes = np.array(self.hidden_layer_sizes).astype(int)
        super()._fit(X, y, incremental=False)


class MLPClassifier_Wrapper(MLPClassifier):

    def _fit(self, X, y, incremental=False):

        self.hidden_layer_sizes = np.array(self.hidden_layer_sizes).astype(int)
        super()._fit(X, y, incremental=False)



from ..BPtSearchCV import (wrap_param_search, get_search_cv,
                           NevergradSearchCV, BPtGridSearchCV)
from ...default.params.Params import Choice
from .helpers import get_param_search


def test_wrap_param_search():

    param_search = get_param_search()
    model_obj = ('fake_name', 'fake_model')
    model_params = {'fake_name__param1': 1,
                    'fake_name__param2': Choice([1, 2]),
                    'other_model__param1': 1}

    search_obj, search_params =\
        wrap_param_search(param_search, model_obj, model_params)

    # Assert is name, exact name might change
    # but the original name should be present
    search_name = search_obj[0]
    assert isinstance(search_name, str)
    assert 'fake_name' in search_name

    search_model = search_obj[1]
    assert isinstance(search_model, NevergradSearchCV)
    assert(search_model.estimator == 'fake_model')

    # Should return other models params
    assert len(search_params) == 1
    assert search_params['other_model__param1'] == 1

    # Check search params
    params = search_model.param_distributions
    assert params['param1'] == 1
    assert len(params['param2']) == 2

    assert search_model.n_jobs == 2


def test_get_search_cv():

    param_search = get_param_search()
    estimator = 'fake_estimator'
    param_distributions = {'param1': 5}
    param_search['progress_loc'] = 'nowhere'

    search_model = get_search_cv(estimator, param_search,
                                 param_distributions)

    assert isinstance(search_model, NevergradSearchCV)
    assert(search_model.estimator == 'fake_estimator')
    assert search_model.n_jobs == 2
    assert search_model.param_distributions['param1'] == 5
    assert search_model.ps['progress_loc'] == 'nowhere'

    # Make sure get params works
    params = search_model.get_params()
    assert params['estimator'] == 'fake_estimator'
    assert params['ps']['n_jobs'] == 2
    assert params['ps']['random_state'] == 1


def test_get_grid_search():

    param_search = get_param_search()
    param_search['search_type'] = 'grid'
    estimator = 'fake_estimator'
    param_distributions = {'param1': 5}
    param_search['progress_loc'] = 'nowhere'

    search_model = get_search_cv(estimator, param_search,
                                 param_distributions)

    assert isinstance(search_model, BPtGridSearchCV)



from ...main.input import Imputer, Model
from ..constructors import ImputerConstructor
from ...dataset.Dataset import Dataset
import numpy as np
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from ..ScopeObjs import ScopeTransformer


def test_iterative_imputer():

    spec = {'problem_type': 'binary',
            'random_state': 2,
            'n_jobs': 1,
            'scope': 'all'}

    dataset = Dataset()
    dataset['float col'] = [1, 2, 3, 4, 5, 6]
    dataset['cat'] = [1, 1, 2, 2, 3, 3]
    dataset['cat missing'] = [1, 1, np.nan, 2, np.nan, 3]
    dataset.add_scope(['cat', 'cat missing'], 'category', inplace=True)

    assert dataset.get_cols('category') == ['cat', 'cat missing']

    piece = ImputerConstructor(dataset=dataset, spec=spec, user_passed_objs={})

    input_piece = Imputer('iterative', scope='category',
                          base_model=Model('linear'),
                          base_model_type='default')

    objs, obj_params = piece.process([input_piece])

    assert len(obj_params) == 0

    base_trans = objs[0][1]
    assert isinstance(base_trans, ScopeTransformer)

    assert base_trans.inds == dataset._get_data_inds('all', 'category')

    # Since inds are based on sorted col names
    # we know this should be 0, 1
    assert base_trans.inds == [0, 1]

    it_imputer = base_trans.estimator
    assert isinstance(it_imputer, IterativeImputer)
    assert it_imputer.random_state == 2

    assert it_imputer.estimator.inds is Ellipsis

    # Should be categorical, so logistic regression
    base_estimator = it_imputer.estimator.estimator
    assert isinstance(base_estimator, LogisticRegression)


def test_iterative_imputer_default_float():

    spec = {'problem_type': 'binary',
            'random_state': 2,
            'n_jobs': 1,
            'scope': 'all'}

    dataset = Dataset()
    dataset['float col'] = [1, 2, 3, 4, 5, 6]
    dataset['cat'] = [1, 1, 2, 2, 3, 3]
    dataset['cat missing'] = [1, 1, np.nan, 2, np.nan, 3]
    dataset.add_scope(['cat', 'cat missing'], 'category', inplace=True)

    assert dataset.get_cols('category') == ['cat', 'cat missing']

    piece = ImputerConstructor(dataset=dataset, spec=spec, user_passed_objs={})

    input_piece = Imputer('iterative', scope='all',
                          base_model=Model('linear'),
                          base_model_type='default')

    objs, obj_params = piece.process([input_piece])

    assert len(obj_params) == 0

    base_trans = objs[0][1]
    assert isinstance(base_trans, ScopeTransformer)
    assert base_trans.inds is Ellipsis

    it_imputer = base_trans.estimator
    assert isinstance(it_imputer, IterativeImputer)
    assert it_imputer.random_state == 2

    assert it_imputer.estimator.inds is Ellipsis

    # Not all categorical, so should be linear regression
    base_estimator = it_imputer.estimator.estimator
    assert isinstance(base_estimator, LinearRegression)



from .Params import Scalar

P = {}

# Transformers
P['pca var search'] =\
        {'n_components': Scalar(init=.75, lower=.1, upper=.99),
         'svd_solver': 'full'}

P['ohe'] =\
        {'sparse': False,
         'handle_unknown': 'ignore'}

P['dummy code'] =\
        {'sparse': False,
         'drop': 'first',
         'handle_unknown': 'error'}



from ..main.input import (Ensemble, FeatSelector, Scaler, Pipeline, Model,
                          ParamSearch, Transformer, Imputer)
from ..main.compare import Compare, Option


# Pieces
m_imputer = Imputer('mean', scope='float')
c_imputer = Imputer('median', scope='category')
r_scaler = Scaler('robust', scope='float')
ohe = Transformer('one hot encoder', scope='category')

random_search = ParamSearch('RandomSearch', n_iter=60)
elastic_search = Model('elastic', params=1, param_search=random_search)
lgbm_search = Model('lgbm', params=1, param_search=random_search)
ridge_search = Model('ridge', params=1, param_search=random_search)

u_feat = FeatSelector('univariate selection', params=2)
svm = Model('svm', params=1)
svm_search_pipe = Pipeline(steps=[u_feat, svm], param_search=random_search)
svm_search = Model(svm_search_pipe)

stacking = Ensemble('stacking', models=[elastic_search,
                                        ridge_search,
                                        svm_search,
                                        lgbm_search],
                    base_model=ridge_search,
                    n_jobs_type='models')

# Pre-defined pipelines
_base_steps = [m_imputer, c_imputer, r_scaler, ohe]

elastic_pipe = Pipeline(steps=_base_steps + [elastic_search])
ridge_pipe = Pipeline(steps=_base_steps + [ridge_search])
lgbm_pipe = Pipeline(steps=_base_steps + [lgbm_search])
svm_pipe = Pipeline(steps=_base_steps + [svm_search])
stacking_pipe = Pipeline(steps=_base_steps + [stacking])

compare_pipe = Compare([Option(elastic_pipe, name='elastic'),
                        Option(ridge_pipe, name='ridge'),
                        Option(svm_pipe, name='svm'),
                        Option(lgbm_pipe, name='lgbm')])

pieces = {'m_imputer': m_imputer,
          'c_imputer': c_imputer,
          'r_scaler': r_scaler,
          'ohe': ohe,
          'random_search': random_search,
          'elastic_search': elastic_search,
          'lgbm_search': lgbm_search,
          'u_feat': u_feat,
          'svm': svm,
          'svm_search_pipe': svm_search_pipe,
          'svm_search': svm_search,
          'ridge_search': ridge_search,
          'stacking': stacking}

pipelines = {'elastic_pipe': elastic_pipe,
             'ridge_pipe': ridge_pipe,
             'lgbm_pipe': lgbm_pipe,
             'svm_pipe': svm_pipe,
             'stacking_pipe': stacking_pipe,
             'compare_pipe': compare_pipe}

pieces_keys = list(pieces)
pipelines_keys = list(pipelines)



import pandas as pd
from ..residualizer import LinearResidualizer
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression
import numpy as np


def test_base():

    covars = pd.DataFrame()

    covars['1'] = [1, 1, 1]
    covars['2'] = [1, 2, 3]
    covars['2'] = covars['2'].astype('category')

    model = LinearResidualizer(to_resid_df=covars,
                               demean=True,
                               dummy_code=True)

    X = np.ones((3, 5))
    model.fit(X, fit_index=[0, 1, 2])

    assert len(model.means_) == 3

    # Mean of all 1's col should be 1
    assert model.means_[0] == 1
    assert isinstance(model.encoder_, OneHotEncoder)
    assert isinstance(model.estimator_, LinearRegression)

    Xt = model.transform(X, transform_index=[0, 1, 2])
    assert Xt.shape == (3, 5)


def test_with_nans():

    covars = pd.DataFrame()

    covars['1'] = [1, 1, 1]
    covars['2'] = [1, 2, 3]
    covars['2'] = covars['2'].astype('category')

    model = LinearResidualizer(to_resid_df=covars,
                               demean=True,
                               dummy_code=True)

    X = np.ones((3, 5))
    X[0][0] = np.nan
    model.fit(X, fit_index=[0, 1, 2])

    assert len(model.means_) == 3

    # Mean of all 1's col should be 1
    assert model.means_[0] == 1
    assert isinstance(model.encoder_, OneHotEncoder)
    assert len(model.estimators_) == 5

    Xt = model.transform(X, transform_index=[0, 1, 2])
    assert Xt.shape == (3, 5)



from ..util import conv_to_list, save_docx_table, get_top_substrs
import pandas as pd
import tempfile
import os


def test_conv_to_list_None():

    assert conv_to_list(None) is None


def test_save_docx_table():

    loc = os.path.join(tempfile.gettempdir(), 't.docx')

    df = pd.DataFrame(['1', '2', '3'], columns=['1'])
    save_docx_table(df, loc)

    assert os.path.exists(loc)
    os.remove(loc)


def test_get_top_substrs():

    e = get_top_substrs(['apple_123', 'dog_123'])
    assert len(e) == 1
    assert e[0] == '_123'


def test_get_top_substrs_no_overlap():

    e = get_top_substrs(['n', 'g', 'x'])
    assert len(e) == 0



import numpy as np
from sklearn.preprocessing import LabelEncoder, KBinsDiscretizer
from .helpers import (add_new_categories, remove_unused_categories,
                      get_str_round)
from operator import add
from functools import reduce
from .Dataset import _shared_docs, _sip_docs
from pandas.util._decorators import doc


def _add_new_copy(self, old, new):

    # Copy as new key
    self[new] = self[old].copy()

    # Try to copy scope and role too
    try:
        self.scopes[new] = self.scopes[old].copy()
    except KeyError:
        pass

    try:
        self.roles[new] = self.roles[old]
    except KeyError:
        pass


@doc(**_sip_docs)
def to_binary(self, scope, drop=True, inplace=False):
    '''This method works by setting all
    columns within scope to just two binary
    categories. This works by setting the two values
    as the top two categories, any others will have
    their subjects either dropped or replaced with NaN.

    This method is designed for converting from already
    categorical data to explicitly type binary.

    Parameters
    ----------
    {scope}

    drop : bool, optional
        If set to True, default, then if more
        than two categories are found when converting
        a column to binary, then the subjects / rows
        with these extra values will be dropped from
        the Dataset. If False, then these values will
        be set to NaN and no rows dropped.

        ::

            default = True

    {inplace}

    See Also
    ---------
    binarize : For converting float data to binary.


    Notes
    ----------
    This function with not work on columns
    of type :ref:`data_files`.


    Examples
    ---------
    Simple example with drop True and False below:

    .. ipython:: python

        data = bp.read_csv('data/example1.csv')
        data

        data.to_binary('all', drop=True)
        data.to_binary('all', drop=False)
    '''

    if not inplace:
        return self._inplace('to_binary', locals())

    # Make sure encoders init'ed
    self._check_encoders()

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # Check for data files
    self._data_file_fail_check(cols)

    # Binarize each
    for col in cols:
        self._base_binarize(col=col, drop=drop)

    # Update scopes
    self._check_scopes()


def _base_binarize(self, col, drop):

    # Extract non-nan values / series
    values = self._get_values(col, dropna=True)

    # Get value counts
    value_counts = values.value_counts()

    # If only 1 values
    if len(value_counts) == 1:
        self._print('Warning: to_binary ' + repr(col) + ' was '
                    'passed with only 1 unique value.')

    # Assuming should be binary, so 2 unique values
    if len(value_counts) > 2:

        # Select all but top two by count to drop
        drop_vals = value_counts.sort_values(ascending=False).index[2:]

        # Get to drop
        to_drop = values.index[values.isin(drop_vals)]

        # Drop or NaN
        if drop:
            self.drop(list(to_drop), inplace=True)
        else:
            self.loc[to_drop, col] = np.nan

    # Ordinalize
    self._ordinalize(col)


@doc(**_sip_docs)
def binarize(self, scope, threshold, replace=True, drop=True, inplace=False):
    '''This method contains a utilities for binarizing a variable.
    These are dichotomizing an existing variable with parameter
    threshold, and applying binarization via two thresholds
    (essentially chopping out the middle of the distribution).

    Parameters
    ----------
    {scope}

    threshold : float or (float, float)
        This parameter can be used to either set a single threshold
        where any values less than or equal (>=) to the threshold
        will be set to 0, and any values greater (<) than the threshold
        will be set to 1.

        Alternatively, in the case that a tuple with two
        values is passed, e.g., (5, 10), then this requests
        that a lower and upper threshold be set, with any
        values in the middle either dropped or set to NaN
        (as dependent on the drop parameter). The first
        value of the tuple represents the lower threshold,
        where any values less than this threshold will be set to
        0. The second element of the tuple represents the upper
        threshold where any values greater than this threshold will
        be set to 1. Note these equalities are strictly less than
        or greater than, e.g., not less than or equal to.

    replace : bool, optional
        This parameter controls if the original columns
        should be replaced with their binary version, when
        set to True, and if set to False will add a new
        binary column as well as leave the original column.
        The new columns will share the name of the original
        columns but with '_binary' appended.

        ::

            default = True

    drop : bool, optional
        If set to True, then any values between lower and upper
        will be dropped. If False, they will be set to NaN.

        Note: This parameter is only relevant if using
        the configuration with parameter's upper and lower.

        ::

            default = True

    {inplace}

    See Also
    ---------
    to_binary : Convert from categorical to binary.


    Notes
    ----------
    This function with not work on columns
    of type :ref:`data_files`.

    Examples
    ----------

    .. ipython:: python

        data = bp.read_csv('data/example2.csv')
        data

        data.binarize(scope='all', threshold=.5)
        data.binarize(scope='all', threshold=(.3, .6), drop=False)


    '''

    if not inplace:
        return self._inplace('binarize', locals())

    # Proc if tuple or not
    if isinstance(threshold, tuple):
        if len(threshold) != 2:
            raise RuntimeError('If a tuple passed for threshold it '
                               'must be of length 2!')

        lower, upper = threshold
        threshold = None

    # Not tuple case
    else:
        lower, upper = None, None

    # Make sure encoders init'ed
    self._check_encoders()

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # Check for data files
    self._data_file_fail_check(cols)

    # Binarize each column individually
    for col in cols:
        self._binarize(col=col, threshold=threshold, lower=lower,
                       upper=upper, replace=replace, drop=drop)

    # Finalize scope
    self._check_scopes()


def _binarize(self, col, threshold, lower, upper, replace, drop):

    # Keep track of initial NaN subjects
    nan_subjects = self._get_nan_subjects(col)

    # Extract non-nan values / series
    values = self._get_values(col, dropna=True)

    # Add new column
    if not replace:
        new_key = col + '_binary'

        if new_key in self.columns:
            cnt = 1
            while new_key + str(cnt) in self.columns:
                cnt += 1

            new_key += str(cnt)

        # Make a copy under the new key
        self._add_new_copy(old=col, new=new_key)

        # Change col to new_key
        col = new_key

    # If upper and lower passed instead of threshold
    if threshold is None:

        # Determine subjects that are between the values
        to_drop = values[(values <= upper) &
                         (values >= lower)].index

        # If drop
        if drop:
            self.drop(to_drop, axis=0, inplace=True)

        # Grab copy of column to fill
        # after drops
        new_col = self[col].copy()

        # Binarize
        new_col = new_col.where(new_col > lower, 0)
        new_col = new_col.where(new_col < upper, 1)

        # If not dropping, replace as NaN here
        if not drop:
            new_col.loc[to_drop] = np.nan

        # Make sure column is now categorical and replace
        new_col = new_col.astype("Int64").astype('category')
        remove_unused_categories(new_col)
        self[col] = new_col

        # Add to name map
        self.encoders[col] =\
            {0: '<' + str(lower), 1: '>' + str(upper)}

    # If a single threshold
    else:

        # Get copy of column to operate on
        new_col = self[col].copy()

        # Binarize
        new_col = new_col.where(new_col >= threshold, 0)
        new_col = new_col.where(new_col < threshold, 1)

        # Make sure column is now categorical and replace
        new_col = new_col.astype("Int64").astype('category')
        remove_unused_categories(new_col)
        self[col] = new_col

        # Add to name map
        self.encoders[col] =\
            {0: '<' + str(threshold), 1: '>=' + str(threshold)}

    # Make sure NaN's are filled in
    self.loc[nan_subjects, col] = np.nan

    # Make sure category scope
    self._add_scope(col, 'category')


@doc(**_sip_docs)
def k_bin(self, scope, n_bins=5, strategy='uniform', inplace=False):
    '''This method is used to apply k binning to
    a column, or columns. On the backend
    this function used the scikit-learn
    KBinsDiscretizer.

    Parameters
    ----------
    {scope}

    n_bins : int, optional
        The number of bins to discretize the passed
        columns to. This same value is applied for
        all columns within scope.

        ::

            default = 5

    strategy : 'uniform', 'quantile' or 'kmeans', optional
        The strategy in which the binning should be adhere to.
        Options are:

        - 'uniform'
            All bins in each feature have identical widths.

        - 'quantile'
            All bins in each feature have the same number of points.

        - 'kmeans'
            Values in each bin have the same nearest center of a 1D
            k-means cluster.

        ::

            default = 'uniform'

    {inplace}

    Examples
    ---------
    .. plot::
        :context: close-figs

        import BPt as bp
        data = bp.Dataset([.1, .2, .3, .4, .5, .6, .7, .8, .9],
                          columns=['feat'])

        # Apply k_bin, not in place, then plot
        data.k_bin('feat', n_bins=3, strategy='uniform').plot('feat')

        # Apply with dif params
        data.k_bin('feat', n_bins=5, strategy='quantile').plot('feat')

    '''

    if not inplace:
        return self._inplace('k_bin', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # K bin each requested column individually
    for col in cols:
        self._k_bin(col, n_bins, strategy)

    # Make sure scope updated
    self._check_scopes()


def _k_bin(self, col, n_bins, strategy):

    # Check encoders init'ed
    self._check_encoders()

    # Get non-nan values
    all_values = self[col]
    non_nan_subjects = all_values[~all_values.isna()].index
    values = all_values.loc[non_nan_subjects]

    # Prepare values for encoding
    values = np.array(values.astype('float')).reshape(-1, 1)

    # Perform encoding
    kbin_encoder = KBinsDiscretizer(n_bins=n_bins,
                                    encode='ordinal',
                                    strategy=strategy)
    new_values = np.squeeze(kbin_encoder.fit_transform(values))

    self._print('k_bin encoded col:', col, 'with:',
                kbin_encoder.n_bins, 'bins and strategy:',
                strategy, level=1)

    # Add any needed new categories to the column
    self._replace_cat_values(col, new_values, non_nan_subjects)

    # Create encoder information
    encoder = {}
    bin_edges = kbin_encoder.bin_edges_[0]

    for i in range(len(bin_edges)-1):

        # Get rounded start and end values
        start = get_str_round(bin_edges[i], places=3)
        end = get_str_round(bin_edges[i+1], places=3)

        # Set ordinal index to str name of range values
        encoder[i] = start + ' - ' + end

    # Save in encoders
    self.encoders[col] = encoder


@doc(**_sip_docs)
def ordinalize(self, scope, nan_to_class=False, inplace=False):
    '''This method is used to ordinalize
    a group of columns. Ordinalization is
    performed by setting all n unique
    categories present in each column to
    values 0 to n-1.

    The :class:`LabelEncoder<sklearn.preprocessing.LabelEncoder>`
    is used on the backend for this operation.

    Parameters
    -----------
    {scope}

    nan_to_class : bool, optional
        | If set to True, then treat NaN values as
          as a unique class, otherwise if False then
          ordinalization will be applied on just non-NaN
          values, and any NaN values will remain NaN.

        | See: :func:`nan_to_class <Dataset.nan_to_class>`
          for more generally adding NaN values as a new
          category to any arbitrary categorical column.

        ::

            default = False

    {inplace}

    Examples
    ---------

    .. ipython:: python

        data = bp.read_csv('data/example1.csv')
        data

        data = data.ordinalize('all')
        data

    Note that the original names are still saved when using
    this and simmilar encoding functions.

    .. ipython:: python

        data.encoders

    '''

    if not inplace:
        return self._inplace('ordinalize', locals())

    # Check scope and role
    self._check_sr()

    # Get cols from scope
    cols = self._get_cols(scope)

    # Ordinalize each column individually
    for col in cols:
        self._ordinalize(col)

    # Optionally add NaN as class
    if nan_to_class:
        self.nan_to_class(scope=scope, inplace=True)

    # Make sure scope updated
    self._check_scopes()


def _ordinalize(self, col):

    # Check encoders init'ed
    self._check_encoders()

    # Get non-nan values
    all_values = self[col]
    non_nan_subjects = all_values[~all_values.isna()].index
    values = all_values.loc[non_nan_subjects]

    # Perform actual encoding
    label_encoder = LabelEncoder()
    new_values = label_encoder.fit_transform(np.array(values))

    # Add any needed new categories to the column
    self._replace_cat_values(col, new_values, non_nan_subjects)

    # Convert label encoder to encoder
    encoder = {}
    for i, c in enumerate(label_encoder.classes_):
        encoder[i] = c

    # Save in encoders
    self.encoders[col] = encoder


def _replace_cat_values(self, col, new_values, non_nan_subjects):

    # Make copy of column
    new_col = self[col].copy()

    # Add new categories if needed / already categorical
    add_new_categories(new_col, new_values)
    new_col.loc[non_nan_subjects] = new_values

    # Make sure cast to type nullable integer, then
    # category.
    new_col = new_col.astype("Int64").astype('category')
    remove_unused_categories(new_col)

    # Add back in place
    self[col] = new_col

    # Make sure col is category type
    self._add_scope(col, 'category')


@doc(**_sip_docs)
def nan_to_class(self, scope='category', inplace=False):
    '''This method will cast any columns that were not categorical that are
    passed here to categorical. Will also ordinally encode them if
    they have not already been encoded
    (e.g., by either via ordinalize, binarize, or a similar function...).

    {scope}
        ::

            default = 'category'

    {inplace}
    '''

    if not inplace:
        return self._inplace('nan_to_class', locals())

    # Check scope and roles
    self._check_sr()

    # Make sure encoders are up init'ed
    self._check_encoders()

    # For each col in scope
    cols = self._get_cols(scope)
    for col in cols:

        # If this column hasn't been encoded yet, ordinalize
        if col not in self.encoders:
            self._ordinalize(col)

        # Get nan subjects
        all_values = self[col]
        nan_subjects = all_values[all_values.isna()].index

        # If no nan subjects, skip
        if len(nan_subjects) == 0:
            continue

        # Get non-nan values
        non_nan_subjects = all_values[~all_values.isna()].index
        values = all_values.loc[non_nan_subjects]

        # Add nan class as next avaliable
        nan_class = np.max(values.astype(int)) + 1

        # Get a copy of the new column
        new_col = self[col].copy()

        # Add categories, replace vals, and remove unused
        add_new_categories(new_col, [nan_class])
        new_col.loc[nan_subjects] = nan_class
        remove_unused_categories(new_col)

        # Replace col with new column
        self[col] = new_col

        # Update encoder entry with NaN
        self.encoders[col][nan_class] = np.nan

    # Make sure scopes updated
    self._check_scopes()


@doc(inplace=_shared_docs['inplace'])
def copy_as_non_input(self, col, new_col, copy_scopes=True, inplace=False):
    '''This method is a used for making a copy of an
    existing column, ordinalizing it and then setting it
    to have role = non input.

    Parameters
    ----------
    col : str
        The name of the loaded column to make a copy of.

    new_col : str
        The new name of the non input and ordinalized column.

    copy_scopes : bool, optional
        | If the associated scopes with the original column should be copied
          over as well. If False, then the new col will
          only have scope 'category'.

        | The scopes will be copied before ordinalizing,
          such that the new copy will have the scope 'category'
          regardless of if that was a scope of the original variable.

        ::

            default = True

    {inplace}

    '''

    if not inplace:
        return self._inplace('copy_as_non_input', locals())

    # Copy as new col
    self[new_col] = self[col].copy()

    # Update scope and role
    self._check_sr()

    # Only copy scopes if requested, and do before
    # ordinalize call
    if copy_scopes:
        self.scopes[new_col] = self.scopes[col].copy()

    # Ordinalize
    self._ordinalize(new_col)

    # Set new role
    self.set_role(new_col, 'non input', inplace=True)

    # Make sure scopes updated
    self._check_scopes()


@doc(inplace=_shared_docs['inplace'])
def add_unique_overlap(self, cols, new_col, decode_values=True, inplace=False):
    '''| This function is designed to add a new column
      with the overlapped unique values from passed two or more columns.

    | The new added column will be default be added with role data,
      except if all of the passed cols have a different role. In the
      case that all of the passed cols have the same role, the new
      col will share that role. Simmilar to role, the scope of
      the new column will be the overlap
        of shared scopes from all of the passed new_col. If no overlap,
        then no scope.

    Parameters
    -----------
    cols : list of str
        The names of the columns to compute the overlap with.
        E.g., in the example above, cols = ['A', 'B'].

        Note: You must pass atleast two columns here.

    new_col : str
        The name of the new column where these values will be stored.

    decode_values : bool, optional
        This is an optional parameter, which is set to True
        will when creating the overlapping values will try to
        replace values with the encoded value (if any). For example
        if a variable being added had an originally encoded values of
        'cat' and 'dog', then the replace value before ordinalization
        would be col_name=cat and col_name=dog, vs. if set to False
        would have values of col_name=0 and col_name=1.

        ::

            default = True

    {inplace}


    Examples
    -----------

    .. ipython:: python

        data = bp.read_csv('data/example1.csv')
        data

        data.add_unique_overlap(cols=['animals', 'numbers'],
                                new_col='combo', inplace=True)
        data
        data.encoders['combo']

    In that example every combination was a unique combination.
    Let's try again, but now with overlaps.

    .. ipython:: python

        data = bp.read_csv('data/example1.csv')
        data = data.ordinalize('all')

        data.add_unique_overlap(cols=['animals', 'numbers'],
                                new_col='combo', inplace=True)
        data
        data.encoders['combo']

    '''

    if not inplace:
        return self._inplace('add_unique_overlap', locals())

    # Make sure up to date
    self._check_sr()

    # Some input validation, make sure str
    if isinstance(cols, str):
        raise RuntimeError('You must pass cols as a list or array-like.')

    # Make sure 2 or more cols
    if len(cols) < 2:
        raise RuntimeError('cols must be of length 2 or more.')

    # Make sure passed cols exist
    for col in cols:
        if col not in list(self):
            raise KeyError('Passed col: ' + col + ' is not a valid '
                           'loaded column.')

    # Make sure new col doesn't exist
    if new_col in list(self):
        raise KeyError('Passed new col: ' + new_col + ' already exists!')

    # Generate a list of modified series
    combo = []
    for col in cols:

        vals = self._get_values(col, dropna=False,
                                decode_values=decode_values)
        combo.append(col + '=' + vals.astype(str) + ' ')

    # Combine
    combo = reduce(add, combo)

    # Add as new column
    self[new_col] = combo

    # Update scope and role
    self._check_sr()

    # If all roles agree, set new col as
    roles = set([self.roles[col] for col in cols])
    if len(roles) == 1:
        self._set_role(new_col, roles.pop())

    # Add scope if in all passed
    scopes = [self.scopes[col] for col in cols]
    for scope in scopes[0]:
        in_rest = [scope in other for other in scopes[1:]]
        if all(in_rest):
            self._add_scope(new_col, scope)

    # Lastly, ordinalize
    self._ordinalize(new_col)

    # Make sure scopes updated
    self._check_scopes()



from ...pipeline.helpers import proc_mapping
from ...pipeline.BPtTransformer import BPtTransformer
from sklearn.utils.validation import check_memory


def _fit_estimator(estimator, Xs, y=None, **fit_params):

    estimator.fit(Xs=Xs, y=y, **fit_params)
    return estimator


class BPtTransformerMV(BPtTransformer):

    def _proc_mapping(self, mapping):
        '''self.inds_ will be the same for compat. but view_inds_
        will be used for fit.'''

        # Save a copy of the passed mapping
        self.mapping_ = mapping.copy()

        self.inds_ = []
        for inds in self.inds:
            self.inds_ += proc_mapping(inds, mapping)
        self.inds_ = sorted(list(set(self.inds_)))

        # Calculate view inds seperately
        self.view_inds_ = [proc_mapping(inds, mapping)
                           for inds in self.inds]

        return self

    def _fit(self, X, y=None, **fit_params):
        '''Override this function from ScopeObjs parent class almost exactly
        as is, but passing Xs instead of X.'''

        # Get correct fit function as either with memory
        # caching, or just as is, if no cache loc passed.
        if self.cache_loc is not None:
            memory = check_memory(self.cache_loc)
            _fit_estimator_c = memory.cache(_fit_estimator)
        else:
            _fit_estimator_c = _fit_estimator

        # Fit the estimator
        self.estimator_ =\
            _fit_estimator_c(estimator=self.estimator_,
                             Xs=[X[:, inds] for inds in self.view_inds_],
                             y=y, **fit_params)

    def _est_transform(self, X, **trans_params):

        return self.estimator_.transform(
            Xs=[X[:, inds] for inds in self.view_inds_], **trans_params)



from ..BPtLGBM import BPtLGBMClassifier, BPtLGBMRegressor
from ...main.CV import BPtCV, CVStrategy
import numpy as np


def test_basic():

    X = np.ones((20, 20))
    y = np.ones((20))
    y[:10] = 0

    # Just shouldn't fail
    regr = BPtLGBMRegressor()
    regr.fit(X, y)
    regr.predict(X)

    clasif = BPtLGBMClassifier()
    clasif.fit(X, y)
    clasif.predict(X)


def test_with_bpt_cv():

    cv = BPtCV(splits=.5, n_repeats=1, cv_strategy=CVStrategy(),
               splits_vals=None, random_state=1)

    X = np.ones((20, 20))
    y = np.ones((20))

    # Make sure fit works w/ custom CV
    regr = BPtLGBMRegressor(eval_split=cv, early_stopping_rounds=10)
    regr.fit(X, y, fit_index=np.arange(20))

    X_eval, y_eval, eval_set =\
        regr._get_eval_set(X, y, fit_index=np.arange(20))

    assert X_eval.shape == (10, 20)
    assert y_eval.shape == (10, )
    assert eval_set[0].shape == (10, 20)
    assert eval_set[1].shape == (10, )

    # Regressor w/o early stop rounds
    regr = BPtLGBMRegressor(eval_split=cv, early_stopping_rounds=None)
    X_eval, y_eval, eval_set =\
        regr._get_eval_set(X, y, fit_index=np.arange(20))

    assert X_eval.shape == (20, 20)
    assert eval_set is None


def test_cv_as_size():

    X = np.ones((20, 20))
    y = np.ones((20))

    # Check works with cv as size
    regr = BPtLGBMRegressor(eval_split=.5, early_stopping_rounds=10)
    regr.fit(X, y, fit_index=np.arange(20))

    X_eval, y_eval, eval_set =\
        regr._get_eval_set(X, y, fit_index=np.arange(20))

    assert X_eval.shape == (10, 20)
    assert y_eval.shape == (10, )
    assert eval_set[0].shape == (10, 20)
    assert eval_set[1].shape == (10, )


def test_with_cat_features():

    X = np.ones((5, 5))
    y = np.ones((5))
    base_mapping = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}

    regr = BPtLGBMRegressor(cat_inds=[0, 1])
    regr.fit(X, y)

    categorical_feature = regr._get_categorical_feature(base_mapping)
    assert categorical_feature == [0, 1]

    mapping = {0: [0, 1, 2], 1: [1, 2, None], 2: None, 3: 3, 4: 4, 5: [1, 2]}
    categorical_feature = regr._get_categorical_feature(mapping)
    assert categorical_feature == [0, 1, 2]




from copy import deepcopy
from .helpers import set_n_jobs, replace_with_in_params

from sklearn.ensemble import (StackingRegressor, StackingClassifier,
                              VotingClassifier, VotingRegressor)

from joblib import Parallel, delayed
from sklearn.base import clone, is_classifier
from sklearn.utils import Bunch
from sklearn.model_selection import check_cv, cross_val_predict
import numpy as np
from .base import _fit_single_estimator, _get_est_fit_params
from ..main.CV import BPtCV

from sklearn.utils.multiclass import check_classification_targets
from sklearn.preprocessing import LabelEncoder
from .helpers import get_mean_fis


def _fit_all_estimators(self, X, y, sample_weight=None, mapping=None,
                        fit_index=None):

    # Validate
    names, all_estimators = self._validate_estimators()

    # Fit all estimators
    self.estimators_ = Parallel(n_jobs=self.n_jobs)(
        delayed(_fit_single_estimator)(clone(est), X, y, sample_weight,
                                       mapping, fit_index)
        for est in all_estimators if est != 'drop'
    )

    self.named_estimators_ = Bunch()
    est_fitted_idx = 0
    for name_est, org_est in zip(names, all_estimators):
        if org_est != 'drop':
            self.named_estimators_[name_est] = self.estimators_[
                est_fitted_idx]
            est_fitted_idx += 1
        else:
            self.named_estimators_[name_est] = 'drop'

    return names, all_estimators


def voting_fit(self, X, y, sample_weight=None, mapping=None,
               fit_index=None, **kwargs):

    # Fit self.estimators_ on all data
    self._fit_all_estimators(
        X, y, sample_weight=sample_weight, mapping=mapping,
        fit_index=fit_index)

    return self


def stacking_fit(self, X, y, sample_weight=None, mapping=None,
                 fit_index=None, **kwargs):

    # Validate final estimator
    self._validate_final_estimator()

    # Fit self.estimators_ on all data
    names, all_estimators = self._fit_all_estimators(
        X, y, sample_weight=sample_weight, mapping=mapping,
        fit_index=fit_index)

    # To train the meta-classifier using the most data as possible, we use
    # a cross-validation to obtain the output of the stacked estimators.

    # If BPtCV call get_cv
    if isinstance(self.cv, BPtCV):

        random_state = None
        if hasattr(self, 'random_state'):
            random_state = self.random_state

        cv_inds = self.cv.get_cv(fit_index,
                                 random_state=random_state,
                                 return_index=True)

    # Otherwise treat as sklearn arg directly
    else:
        cv_inds = self.cv

    # To ensure that the data provided to each estimator are the same, we
    # need to set the random state of the cv if there is one and we need to
    # take a copy.
    cv = check_cv(cv_inds, y=y, classifier=is_classifier(self))
    if hasattr(cv, 'random_state') and cv.random_state is None:
        cv.random_state = np.random.RandomState()

    # Proc stack method
    stack_method = [self.stack_method] * len(all_estimators)

    self.stack_method_ = [
        self._method_name(name, est, meth)
        for name, est, meth in zip(names, all_estimators, stack_method)
    ]

    # Base fit params for sample weight
    sample_weight_params = ({"sample_weight": sample_weight}
                            if sample_weight is not None else None)

    # Get the fit params for each indv estimator
    all_fit_params = [_get_est_fit_params(est, mapping=mapping,
                                          fit_index=fit_index,
                                          other_params=sample_weight_params)
                      for est in all_estimators]

    # Catch rare error - TODO come up with fix
    if X.shape[0] == X.shape[1]:
        raise RuntimeError('Same numbers of data points and ',
                           'features can lead to error.')

    # Make the cross validated internal predictions to train
    # the final_estimator
    predictions = Parallel(n_jobs=self.n_jobs)(
        delayed(cross_val_predict)(clone(est), X, y, cv=deepcopy(cv),
                                   method=meth, n_jobs=self.n_jobs,
                                   fit_params=fit_params,
                                   verbose=self.verbose)
        for est, meth, fit_params in zip(all_estimators,
                                         self.stack_method_,
                                         all_fit_params) if est != 'drop'
    )

    # Only not None or not 'drop' estimators will be used in transform.
    # Remove the None from the method as well.
    self.stack_method_ = [
        meth for (meth, est) in zip(self.stack_method_, all_estimators)
        if est != 'drop'
    ]

    # @TODO make sure train data index is concatenated correctly
    X_meta = self._concatenate_predictions(X, predictions)
    _fit_single_estimator(self.final_estimator_, X_meta, y,
                          sample_weight=sample_weight,
                          mapping=None,
                          fit_index=fit_index)

    return self


def ensemble_classifier_fit(self, X, y,
                            sample_weight=None, mapping=None,
                            fit_index=None, **kwargs):

    check_classification_targets(y)

    # To make compatible with each Voting and Stacking ...
    self._le = LabelEncoder().fit(y)
    self.le_ = self._le

    self.classes_ = self._le.classes_

    return self.bpt_fit(X, self._le.transform(y),
                        sample_weight=sample_weight,
                        mapping=mapping,
                        fit_index=fit_index,
                        **kwargs)


class BPtStackingRegressor(StackingRegressor):
    _needs_mapping = True
    _needs_fit_index = True
    _fit_all_estimators = _fit_all_estimators
    fit = stacking_fit


class BPtStackingClassifier(StackingClassifier):
    _needs_mapping = True
    _needs_fit_index = True
    _fit_all_estimators = _fit_all_estimators
    bpt_fit = stacking_fit
    fit = ensemble_classifier_fit


class BPtVotingRegressor(VotingRegressor):
    _needs_mapping = True
    _needs_fit_index = True
    _fit_all_estimators = _fit_all_estimators
    fit = voting_fit

    @property
    def feature_importances_(self):
        return get_mean_fis(self.estimators_, 'feature_importances_')

    @property
    def coef_(self):
        return get_mean_fis(self.estimators_, 'coef_')


class BPtVotingClassifier(VotingClassifier):
    _needs_mapping = True
    _needs_fit_index = True
    _fit_all_estimators = _fit_all_estimators
    bpt_fit = voting_fit
    fit = ensemble_classifier_fit

    @property
    def feature_importances_(self):
        return get_mean_fis(self.estimators_, 'feature_importances_')

    @property
    def coef_(self):
        return get_mean_fis(self.estimators_, 'coef_')


class EnsembleWrapper():

    def __init__(self, model_params, ensemble_params,
                 _get_ensembler, n_jobs, random_state):

        self.model_params = model_params
        self.ensemble_params = ensemble_params
        self._get_ensembler = _get_ensembler
        self.n_jobs = n_jobs
        self.random_state = random_state

    def _update_model_ensemble_params(self, to_add, model=True, ensemble=True):

        if model:
            new_model_params = {}
            for key in self.model_params:
                new_model_params[to_add + '__' + key] =\
                    self.model_params[key]
            self.model_params = new_model_params

        if ensemble:

            new_ensemble_params = {}
            for key in self.ensemble_params:
                new_ensemble_params[to_add + '__' + key] =\
                    self.ensemble_params[key]
            self.ensemble_params = new_ensemble_params

    def _basic_ensemble(self, models, name, ensemble=False):

        if len(models) == 1:
            return models

        else:
            basic_ensemble = self._get_ensembler(models)
            self._update_model_ensemble_params(name, ensemble=ensemble)

            return [(name, basic_ensemble)]

    def get_updated_params(self):

        self.model_params.update(self.ensemble_params)
        return self.model_params

    def wrap_ensemble(self, models, ensemble, ensemble_params,
                      final_estimator=None,
                      final_estimator_params=None):

        # If no ensemble is passed, return either the 1 model,
        # or a voting wrapper
        if ensemble is None or len(ensemble) == 0:
            return self._basic_ensemble(models=models,
                                        name='Default Voting',
                                        ensemble=True)

        # Otherwise special ensembles
        else:

            # If needs a single estimator, but multiple models passed,
            # wrap in ensemble!
            if ensemble_params.single_estimator:
                se_ensemb_name = 'Single-Estimator Compatible Ensemble'
                models = self._basic_ensemble(models,
                                              se_ensemb_name,
                                              ensemble=False)

            # If no split and single estimator
            if ensemble_params.single_estimator:
                return self._wrap_single(models, ensemble,
                                         ensemble_params.n_jobs_type)

            # Last case is, no split/DES ensemble and also
            # not single estimator based
            # e.g., in case of stacking regressor.
            else:
                return self._wrap_multiple(models, ensemble,
                                           final_estimator,
                                           final_estimator_params,
                                           ensemble_params.n_jobs_type,
                                           ensemble_params.cv)

    def _wrap_single(self, models, ensemble_info, n_jobs_type):
        '''If passed single_estimator flag'''

        # Unpack ensemble info
        ensemble_name = ensemble_info[0]
        ensemble_obj = ensemble_info[1][0]
        ensemble_extra_params = ensemble_info[1][1]

        # Models here since single estimator is assumed
        # to be just a list with
        # of one tuple as
        # [(model or ensemble name, model or ensemble)]
        base_estimator = models[0][1]

        # Set n jobs based on passed type
        if n_jobs_type == 'ensemble':
            model_n_jobs = 1
            ensemble_n_jobs = self.n_jobs
        else:
            model_n_jobs = self.n_jobs
            ensemble_n_jobs = 1

        # Set model / base_estimator n_jobs
        set_n_jobs(base_estimator, model_n_jobs)

        # Make sure random_state is set (should be already)
        if hasattr(base_estimator, 'random_state'):
            setattr(base_estimator, 'random_state', self.random_state)

        # Create the ensemble object
        ensemble = ensemble_obj(base_estimator=base_estimator,
                                **ensemble_extra_params)

        # Set ensemble n_jobs
        set_n_jobs(ensemble, ensemble_n_jobs)

        # Set random state
        if hasattr(ensemble, 'random_state'):
            setattr(ensemble, 'random_state', self.random_state)

        # Wrap as object
        new_ensemble = [(ensemble_name, ensemble)]

        # Have to change model name to base_estimator
        self.model_params =\
            replace_with_in_params(self.model_params, models[0][0],
                                   'base_estimator')

        # Append ensemble name to all model params
        self._update_model_ensemble_params(ensemble_name,
                                           ensemble=False)

        return new_ensemble

    def _wrap_multiple(self, models, ensemble_info,
                       final_estimator, final_estimator_params,
                       n_jobs_type, cv):
        '''In case of no split/DES ensemble, and not single estimator based.'''

        # Unpack ensemble info
        ensemble_name = ensemble_info[0]
        ensemble_obj = ensemble_info[1][0]
        ensemble_extra_params = ensemble_info[1][1]

        # Models here just self.models a list of tuple of
        # all models.
        # So, ensemble_extra_params should contain the
        # final estimator + other params

        # Set model_n_jobs and ensemble n_jobs based on type
        if n_jobs_type == 'ensemble':
            model_n_jobs = 1
            ensemble_n_jobs = self.n_jobs
        else:
            model_n_jobs = self.n_jobs
            ensemble_n_jobs = 1

        # Set the model jobs
        set_n_jobs(models, model_n_jobs)

        # Make sure random state is propegated
        for model in models:
            if hasattr(model[1], 'random_state'):
                setattr(model[1], 'random_state', self.random_state)

        # Determine the parameters to init the ensemble
        pass_params = ensemble_extra_params
        pass_params['estimators'] = models

        # Process final_estimator if passed
        if final_estimator is not None:

            # Replace name of final estimator w/ final_estimator in params
            final_estimator_params =\
                replace_with_in_params(params=final_estimator_params,
                                       original=final_estimator[0][0],
                                       replace='final_estimator')

            # Add final estimator params to model_params - once name changed
            # to avoid potential overlap.
            self.model_params.update(final_estimator_params)

            # Unpack actual model obj
            final_estimator_obj = final_estimator[0][1]

            # Set final estimator n_jobs to model n_jobs
            set_n_jobs(final_estimator_obj, model_n_jobs)

            # Redundant random state check
            if hasattr(final_estimator_obj, 'random_state'):
                setattr(final_estimator_obj, 'random_state', self.random_state)

            # Add to pass params
            pass_params['final_estimator'] = final_estimator_obj

        # Check if cv passed
        if cv is not None:
            pass_params['cv'] = cv

        # Init the ensemble object
        ensemble = ensemble_obj(**pass_params)

        # Set ensemble n_jobs
        set_n_jobs(ensemble, ensemble_n_jobs)

        # Set random state
        if hasattr(ensemble, 'random_state'):
            setattr(ensemble, 'random_state', self.random_state)

        # Wrap as pipeline compatible object
        new_ensemble = [(ensemble_name, ensemble)]

        # Append ensemble name to all model params
        self._update_model_ensemble_params(ensemble_name,
                                           ensemble=False)

        return new_ensemble



"""
Scalers.py
====================================
File containing the various data scalers.
"""
from ..helpers import get_obj_and_params, all_from_objects
from sklearn.preprocessing import (MinMaxScaler, RobustScaler, StandardScaler,
                                   PowerTransformer, MaxAbsScaler,
                                   QuantileTransformer, Normalizer)
from ...extensions.Scalers import Winsorizer


# Scalers differs from models in that the types are not restricted
# by a given problem type. Therefore no AVALIABLE dictionary is needed
# all avaliable scalers are as indexed by SCALERS.
SCALERS = {
    'standard': (StandardScaler, ['base standard']),

    'minmax': (MinMaxScaler, ['base minmax']),

    'maxabs': (MaxAbsScaler, ['default']),

    'robust': (RobustScaler, ['base robust', 'robust gs']),

    'yeo': (PowerTransformer, ['base yeo']),

    'boxcox': (PowerTransformer, ['base boxcox']),

    'winsorize': (Winsorizer, ['base winsorize', 'winsorize gs']),

    'quantile norm': (QuantileTransformer, ['base quant norm']),

    'quantile uniform': (QuantileTransformer, ['base quant uniform']),

    'normalize': (Normalizer, ['default'])
    }


# @TODO find way to re-use these default ones
def get_scaler_and_params(scaler_str, extra_params, params,
                          **kwargs):
    '''Returns a scaler based on proced str indicator input,

    Parameters
    ----------
    scaler_str : str
        `scaler_str` refers to the type of scaling to apply
        to the saved data during model evaluation.

    extra_params : dict
        Any extra params being passed.
        These can be supplied by creating another dict within extra_params.
        E.g., extra_params[method] = {'method param' : new_value}
        Where method param is a valid argument for that method,
        and method in this case is the str indicator.

    params : int
        The index of the params to use.

    Returns
    ----------
    scaler
        A scaler object with fit and transform methods.

    dict
        The params for this scaler
    '''

    scaler, extra_scaler_params, scaler_params =\
        get_obj_and_params(scaler_str, SCALERS, extra_params, params)

    return scaler(**extra_scaler_params), scaler_params


all_obj_keys = all_from_objects(SCALERS)



from .helpers import update_mapping
from .ScopeObjs import ScopeTransformer
from sklearn.preprocessing import OneHotEncoder


class BPtTransformer(ScopeTransformer):

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        # Need the output from a transform to full fit,
        # so when fit is called, call fit_transform instead
        self.fit_transform(X=X, y=y, mapping=mapping,
                           fit_index=fit_index,
                           **fit_params)

        return self

    def _update_transformer_mapping(self, mapping):

        # Need to update the mapping before returning

        # Many to many case for transformer,
        # override existing out_mapping_
        self.out_mapping_ = {}
        X_trans_inds = list(range(self.n_trans_feats_))

        # Many to many case, each ind is mapped
        # to all outputted X_trans_inds
        for i in self.inds_:
            self.out_mapping_[i] = X_trans_inds

        # Fill the remaining spots sequentially,
        # for each of the rest inds.
        for c in range(len(self.rest_inds_)):
            ind = self.rest_inds_[c]
            self.out_mapping_[ind] = self.n_trans_feats_ + c

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

        return self

    def _all_case_update_transformer_mapping(self, X, mapping):

        # Get as list of
        X_trans_inds = list(range(self.n_trans_feats_))

        # All case out mapping
        self.out_mapping_ = {i: X_trans_inds for i in range(X.shape[1])}

        # Since no rest inds, update mapping
        update_mapping(mapping, self.out_mapping_)

        return self

    def fit_transform(self, X, y=None, mapping=None,
                      fit_index=None,
                      transform_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit
        super().fit(X, y=y, mapping=mapping,
                    fit_index=fit_index,
                    **fit_params)

        # If skip
        if self.estimator_ is None:
            return X

        # Transform X - since fit_transform, index is fit index
        X_trans = self.transform(X, transform_index=fit_index)

        # Update mapping and set out_mapping_
        # special all case
        if self.inds_ is Ellipsis:
            self._all_case_update_transformer_mapping(X, mapping)
        else:
            self._update_transformer_mapping(mapping)

        # Now return X_trans
        return X_trans

    def transform_df(self, df, base_name='transformer', encoders=None):
        return super().transform_df(df, base_name=base_name, encoders=encoders)

    def _proc_new_names(self, feat_names, base_name, encoders=None):

        # If skip, return passed names as is
        if self.estimator_ is None:
            return feat_names

        # Check for one hot encoder
        if isinstance(self.estimator_, OneHotEncoder):
            new_names =\
                self._proc_one_hot_new_names(feat_names, encoders=encoders)

        # Get new names
        else:

            if self.inds_ is Ellipsis:
                alt_name = base_name
            elif len(self.inds_) == 1:
                alt_name = feat_names[self.inds_[0]]
            else:
                alt_name = base_name

            try:
                new_names = [alt_name + '_' + str(i)
                             for i in range(self.n_trans_feats_)]
            except IndexError:
                new_names = [base_name + '_' + str(i)
                             for i in range(self.n_trans_feats_)]

        # Remove old names - using parent method
        feat_names = self._remove_old_names(feat_names)

        # New names come first, then rest of names
        all_names = new_names + feat_names

        return all_names

    def _proc_one_hot_new_names(self, feat_names, encoders=None):

        def get_display_cat(name, cat):

            # If encoders passed, and name in encoder
            # use de-coded name as the cat
            if encoders is not None and name in encoders:

                try:
                    cat = encoders[name][cat]

                # If error, keep as is
                except KeyError:
                    cat = cat

            # If Float, turn to int before cast to str
            if isinstance(cat, float):
                cat = int(cat)

            return str(name) + '=' + repr(cat)

        # Scope all case, set inds as identity
        # over all passed feat names
        if self.inds_ is Ellipsis:
            inds = list(range(len(feat_names)))

        # Otherwise use self.inds_
        else:
            inds = self.inds_

        # Save new names in new_names
        new_names = []

        # If no drop
        if self.estimator_.drop_idx_ is None:
            for name_ind, category in zip(inds,
                                          self.estimator_.categories_):
                name = feat_names[name_ind]
                for cat in category:
                    new_names.append(get_display_cat(name, cat))

        # Otherwise if drop index
        else:
            for name_ind, category, to_drop in zip(inds,
                                                   self.estimator_.categories_,
                                                   self.estimator_.drop_idx_):
                name = feat_names[name_ind]
                for i, cat in enumerate(category):
                    if i != to_drop:
                        new_names.append(get_display_cat(name, cat))

        return new_names



from ..BPtModel import BPtModel
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
import numpy as np


def model_test_classes_none():

    model = BPtModel(estimator=DecisionTreeRegressor(), inds=Ellipsis)

    X = np.ones((20, 20))
    y = np.ones((20))

    model.fit(X, y)

    assert model.classes_ is None

    score = model.score(X, y)
    assert score == 1


def model_test_classes_some():

    model = BPtModel(estimator=DecisionTreeClassifier(), inds=Ellipsis)

    X = np.ones((20, 20))
    y = np.ones((20))

    model.fit(X, y)

    assert model.classes_ is not None




from ..Dataset import Dataset
import pandas as pd
from .datasets import (get_fake_dataset, get_fake_dataset2, get_fake_dataset3,
                       get_fake_dataset4, get_fake_dataset6,
                       get_nans_dataset, get_fake_multi_index_dataset)


def test_filter_outliers():

    df = get_fake_dataset()
    df = df.filter_outliers_by_percent(fop=20, scope='3', drop=False)
    assert pd.isnull(df['3']).all()


def test_filter_outliers_inplace():

    df = get_fake_dataset()
    df.filter_outliers_by_percent(fop=20, scope='3',
                                  drop=False, inplace=True)
    assert pd.isnull(df['3']).all()


def test_filter_outliers_by_percent():

    df = get_fake_dataset4()
    df = df.filter_outliers_by_percent(fop=20, scope='1', drop=True)
    assert len(df) == 4

    # Make sure works with NaNs
    df = get_fake_dataset4()
    df = df.filter_outliers_by_percent(fop=20, scope='2', drop=True)
    assert len(df) == 4
    assert pd.isnull(df.loc[5, '2'])

    # Make sure range works
    df = get_fake_dataset4()
    df = df.filter_outliers_by_percent(fop=(20, None), scope='2', drop=True)
    assert len(df) == 5
    assert pd.isnull(df.loc[5, '2'])

    # Make sure drop false works
    df = get_fake_dataset4()
    df = df.filter_outliers_by_percent(fop=(20, None), scope='2', drop=False)
    assert len(df) == 6
    assert pd.isnull(df.loc[0, '2'])
    assert pd.isnull(df.loc[5, '2'])


def test_filter_outliers_by_percent_inplace():

    df = get_fake_dataset4()
    df.filter_outliers_by_percent(fop=20, scope='1', drop=True, inplace=True)
    assert len(df) == 4

    # Make sure works with NaNs
    df = get_fake_dataset4()
    df.filter_outliers_by_percent(fop=20, scope='2', drop=True, inplace=True)
    assert len(df) == 4
    assert pd.isnull(df.loc[5, '2'])

    # Make sure range works
    df = get_fake_dataset4()
    df.filter_outliers_by_percent(fop=(20, None), scope='2',
                                  drop=True, inplace=True)
    assert len(df) == 5
    assert pd.isnull(df.loc[5, '2'])

    # Make sure drop false works
    df = get_fake_dataset4()
    df.filter_outliers_by_percent(fop=(20, None), scope='2',
                                  drop=False, inplace=True)
    assert len(df) == 6
    assert pd.isnull(df.loc[0, '2'])
    assert pd.isnull(df.loc[5, '2'])


def test_filter_outliers_by_std():

    df = get_fake_dataset4()

    # Mean is 3.5, std of 1 is ~1.7 for col 1
    df = df.filter_outliers_by_std(n_std=1, scope='1', drop=True)
    assert len(df) == 4

    # Make sure works with NaNs - mean is 3, std is ~1.4
    df = get_fake_dataset4()
    df = df.filter_outliers_by_std(n_std=1, scope='2', drop=True)
    assert len(df) == 4
    assert pd.isnull(df.loc[5, '2'])

    # Make sure range works
    df = get_fake_dataset4()
    df = df.filter_outliers_by_std(n_std=(1, None), scope='2', drop=True)
    assert len(df) == 5
    assert pd.isnull(df.loc[5, '2'])

    # Make sure drop false works
    df = get_fake_dataset4()
    df = df.filter_outliers_by_std(n_std=(1, None), scope='2', drop=False)
    assert len(df) == 6
    assert pd.isnull(df.loc[0, '2'])
    assert pd.isnull(df.loc[5, '2'])


def test_drop_cols_by_unique_val():

    df = get_fake_dataset2()
    df = df.drop_cols_by_unique_val()

    assert '1' not in df
    assert '2' in df
    assert '3' in df

    df = get_fake_dataset2()
    df = df.drop_cols_by_unique_val(threshold=3)
    assert '1' not in df
    assert '2' not in df
    assert '3' not in df


def test_drop_id_cols():

    df = get_fake_dataset2()
    df = df.drop_id_cols(scope='all')

    assert '1' in df
    assert '2' in df
    assert '3' not in df

    # Inplace = false case
    df = get_fake_dataset2()
    df.drop_id_cols(scope='all', inplace=True)

    assert '1' in df
    assert '2' in df
    assert '3' not in df


def test_drop_duplicate_cols():

    df = get_fake_dataset3()
    df = df.drop_duplicate_cols(scope='all')
    assert '5' in df
    assert df.shape == (3, 3)

    df = get_fake_dataset3()
    df.drop_duplicate_cols(scope='all', inplace=True)
    assert '5' in df
    assert df.shape == (3, 3)


def test_apply_inclusions():

    df = get_fake_dataset3()
    df = df.apply_inclusions([0])
    assert len(df) == 1


def test_apply_exclusions():

    df = get_fake_dataset()
    df = df.apply_exclusions([0, 1])
    assert len(df) == 1

    df = get_fake_dataset()
    df = df.apply_exclusions([0])
    assert len(df) == 2


def test_drop_cols_inclusions():

    df = get_fake_dataset()
    df = df.drop_cols(inclusions='1')
    assert '1' in df
    assert df.shape[1] == 1

    df = get_fake_dataset()
    df = df.drop_cols(inclusions='category')
    assert '2' in df

    df = get_fake_dataset()
    df = df.drop_cols(inclusions=['1', '2'])
    assert df.shape[1] == 2

    df = Dataset(columns=['xxx1', 'xxx2', 'xxx3', '4'])
    df = df.drop_cols(inclusions=['xxx'])
    assert '4' not in df
    assert df.shape[1] == 3


def test_drop_cols_exclusions():

    df = get_fake_dataset()
    df = df.drop_cols(exclusions='1')
    assert '1' not in df
    assert df.shape[1] == 2

    df = get_fake_dataset()
    df = df.drop_cols(exclusions=['1', '2'])
    assert '3' in df
    assert df.shape[1] == 1

    df = Dataset(columns=['xxx1', 'xxx2', 'xxx3', '4'])
    df = df.drop_cols(exclusions=['xxx'])
    assert '4' in df
    assert df.shape[1] == 1


def test_filter_categorical_by_percent():

    df = get_fake_dataset6()
    df.filter_categorical_by_percent(drop_percent=1, scope='all',
                                     drop=True, inplace=True)
    assert len(df) == 10
    assert len(df['2'].unique()) == 3
    assert 'category' in df.scopes['1']
    assert 'category' in df.scopes['2']

    df = get_fake_dataset6()
    df = df.filter_categorical_by_percent(drop_percent=11, scope='all',
                                          drop=True)
    assert len(df) == 9
    assert len(df['2'].unique()) == 2

    df = get_fake_dataset6()
    df = df.filter_categorical_by_percent(drop_percent=11, scope='2',
                                          drop=True)
    assert len(df) == 10

    df = get_fake_dataset6()
    df.filter_categorical_by_percent(drop_percent=20, scope='2',
                                     drop=True, inplace=True)
    assert len(df) == 9
    assert len(df['2'].unique()) == 2

    df = get_fake_dataset6()
    df.filter_categorical_by_percent(drop_percent=20, scope='2',
                                     drop=False, inplace=True)
    assert len(df) == 10
    assert len(df['2'].unique()) == 2
    assert 'category' not in df.scopes['1']
    assert 'category' in df.scopes['2']


def test_drop_subjects_by_nan():

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=1, scope='all')
    assert df.shape == (1, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=.25, scope='all')
    assert df.shape == (1, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=2, scope='all')
    assert df.shape == (2, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=.5, scope='all')
    assert df.shape == (2, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=3, scope='all')
    assert df.shape == (3, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=.75, scope='all')
    assert df.shape == (3, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=4, scope='all')
    assert df.shape == (4, 4)

    df = get_nans_dataset()
    df = df.drop_subjects_by_nan(threshold=.9, scope='all')
    assert df.shape == (4, 4)


def test_drop_cols_by_nan():

    df = get_nans_dataset()
    df = df.drop_cols_by_nan(threshold=1, scope='all')
    assert df.shape == (4, 1)

    df = get_nans_dataset()
    df = df.drop_cols_by_nan(threshold=.25, scope='all')
    assert df.shape == (4, 1)

    df = get_nans_dataset()
    df = df.drop_cols_by_nan(threshold=2, scope='all')
    assert df.shape == (4, 2)

    df = get_nans_dataset()
    df = df.drop_cols_by_nan(threshold=.5, scope='all')
    assert df.shape == (4, 2)

    df = get_nans_dataset()
    df.drop_cols_by_nan(threshold=3, scope='all', inplace=True)
    assert df.shape == (4, 3)

    df = get_nans_dataset()
    df.drop_cols_by_nan(threshold=4, scope='all', inplace=True)
    assert df.shape == (4, 4)

    df = get_nans_dataset()
    df.drop_cols_by_nan(threshold=.9, scope='all', inplace=True)
    assert df.shape == (4, 4)


def test_multi_index_apply_inclusions():

    df = get_fake_multi_index_dataset()
    df = df.apply_inclusions(subjects=['s1'])
    assert df.shape == (2, 3)

    df = get_fake_multi_index_dataset()
    df = df.apply_inclusions(subjects='all')
    assert df.shape == (6, 3)

    df = get_fake_multi_index_dataset()
    df = df.apply_inclusions(subjects=(['s1', 's2'], ['e1']))
    assert df.shape == (2, 3)


def test_multi_index_apply_exclusions():

    df = get_fake_multi_index_dataset()
    df = df.apply_exclusions(subjects=['s1'])
    assert df.shape == (4, 3)

    df = get_fake_multi_index_dataset()
    df = df.apply_exclusions(subjects='all')
    assert df.shape == (0, 3)

    df = get_fake_multi_index_dataset()
    df = df.apply_exclusions(subjects=(['s1', 's2'], ['e1']))
    assert df.shape == (4, 3)



from sklearn.pipeline import Pipeline
import numpy as np
from .helpers import pipe_hash
from joblib import load, dump
import pandas as pd
from sklearn.utils.metaestimators import if_delegate_has_method
import os
from sklearn.utils import _print_elapsed_time
from sklearn.base import clone
from .base import (_get_est_fit_params, _get_est_trans_params,
                   _needs)


# @TODO add docsrings here and also add to docs
class BPtPipeline(Pipeline):

    _needs_mapping = True
    _needs_fit_index = True

    def __init__(self, steps, memory=None,
                 verbose=False,
                 cache_loc=None):

        self.cache_loc = cache_loc
        super().__init__(steps=steps, memory=memory, verbose=verbose)

    @property
    def _needs_transform_index(self):

        # If any steps need it
        for step in self.steps:
            if _needs(step[1], '_needs_transform_index',
                      'transform_index', 'transform'):
                return True

        # Otherwise False
        return False

    @property
    def n_jobs(self):
        '''Return the first n jobs found in steps.
        This function is just meant to be a check to
        see if any of the pipeline steps have n_jobs to set.'''

        for step in self.steps:
            if hasattr(step[1], 'n_jobs'):
                return getattr(step[1], 'n_jobs')

        # Otherwise, return a step we know
        # doesn't have n_jobs
        return self.steps[0][1].n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs):

        # If set here, try to propegate to all steps
        for step in self.steps:
            if hasattr(step[1], 'n_jobs'):
                setattr(step[1], 'n_jobs', n_jobs)

    @property
    def feature_importances_(self):
        if hasattr(self.__getitem__(-1), 'feature_importances_'):
            return getattr(self.__getitem__(-1), 'feature_importances_')
        return None

    @property
    def coef_(self):
        if hasattr(self[-1], 'coef_'):
            return getattr(self[-1], 'coef_')
        return None

    def get_params(self, deep=True):
        params = super()._get_params('steps', deep=deep)
        return params

    def set_params(self, **kwargs):
        super()._set_params('steps', **kwargs)
        return self

    def _fit(self, X, y=None, fit_index=None, **fit_params_steps):

        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()

        # For each transformer
        for (step_idx,
             name,
             transformer) in self._iter(with_final=False,
                                        filter_passthrough=False):

            with _print_elapsed_time('Pipeline',
                                     self._log_message(step_idx)):

                # Skip if passthrough
                if (transformer is None or transformer == 'passthrough'):
                    continue

                # Clone transformer
                cloned_transformer = clone(transformer)

                # Get the correct fit_transform params
                fit_trans_params =\
                    _get_est_fit_params(
                        estimator=cloned_transformer,
                        mapping=self.mapping_,
                        fit_index=fit_index,
                        other_params=fit_params_steps[name],
                        copy_mapping=False)

                # Fit transform the current transformer
                X = cloned_transformer.fit_transform(X=X, y=y,
                                                     **fit_trans_params)

                # Print if an estimator is skipped, if verbose
                if cloned_transformer.estimator_ is None:
                    if self.verbose >= 1:
                        print('Skipping Step:', name, 'due to empty scope.',
                              flush=True)

                # Replace the transformer of the step with the
                # cloned and now fitted transformer
                self.steps[step_idx] = (name, cloned_transformer)

        return X

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        if isinstance(X, pd.DataFrame):

            # Set train data index
            fit_index = X.index

            # Cast to np array
            X = np.array(X)

        if isinstance(y, (pd.DataFrame, pd.Series)):

            # Cast to np array
            y = np.array(y)

        if self.cache_loc is not None:

            # Compute the hash for this fit
            # Store as an attribute
            self.hash_ = pipe_hash([X, y, mapping,
                                   fit_index, fit_params],
                                   self.steps)

            # Check if hash exists - if it does load
            if os.path.exists(self._get_hash_loc()):
                self._load_from_hash()

                # end / return!
                return self

            # Otherwise, continue to fit as normal

        # Set internal mapping as either passed mapping or
        # initialize a new 1:1 mapping.
        if mapping is not None:
            self.mapping_ = mapping.copy()
        else:
            self.mapping_ = {i: i for i in range(X.shape[1])}

        # The base parent fit
        # -------------------

        # Get fit params as indexed by each step
        fit_params_steps = self._check_fit_params(**fit_params)

        # Fit and transform X for all but the last step.
        Xt = self._fit(X, y, fit_index=fit_index,
                       **fit_params_steps)

        # Fit the final step
        with _print_elapsed_time('Pipeline',
                                 self._log_message(len(self.steps) - 1)):
            if self._final_estimator != 'passthrough':

                # Get last params fit params
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]

                # Add mapping and train data index if valid
                fit_params_last_step =\
                    _get_est_fit_params(self._final_estimator,
                                        mapping=self.mapping_,
                                        fit_index=fit_index,
                                        other_params=fit_params_last_step,
                                        copy_mapping=False)

                # Fit the final estimator
                self._final_estimator.fit(Xt, y, **fit_params_last_step)

        # If cache fit enabled, hash fitted pipe here
        if self.cache_loc is not None:
            self._hash_fit()

        return self

    def fit_transform(X, y=None, **fit_params):
        # @ TODO is there a case where we might need this?
        raise RuntimeError('Not currently implemented')

    def _get_hash_loc(self):

        # Make sure directory exists
        os.makedirs(self.cache_loc, exist_ok=True)

        # Set hash loc as directory + hash of fit args
        hash_loc = os.path.join(self.cache_loc, self.hash_)

        return hash_loc

    def _hash_fit(self):

        # Just save full fitted pipeline
        dump(self, self._get_hash_loc())
        return self

    def _load_from_hash(self):

        # Load from saved hash, by
        # loading the fitted object
        # and then copying over
        # each relevant saved fitted parameter
        fitted_pipe = load(self._get_hash_loc())

        # Copy mapping
        self.mapping_ = fitted_pipe.mapping_

        # Copy each step with the fitted version
        for (step_idx,
             name,
             fitted_piece) in fitted_pipe._iter(with_final=True,
                                                filter_passthrough=False):
            self.steps[step_idx] = (name, fitted_piece)

        # Set flag for testing
        self.loaded_ = True

        return self

    def _get_ordered_objs_and_names(self):

        # Get all objects except final model
        ordered_names = [step[0] for step in self.steps[:-1]]
        ordered_objs = [self.__getitem__(name) for name in ordered_names]

        return ordered_objs, ordered_names

    def transform(self, X, transform_index=None):

        Xt = X

        # If DataFrame input
        if isinstance(Xt, pd.DataFrame):

            # Set transform index, then cast to array
            transform_index = Xt.index
            Xt = np.array(Xt)

        # For each transformer, but the last
        for step in self.steps[:-1]:
            transformer = step[1]

            # Get any needed transform params
            trans_params =\
                _get_est_trans_params(transformer,
                                      transform_index=transform_index)

            # Transform X - think in place is okay
            Xt = transformer.transform(Xt, **trans_params)

        return Xt

    def transform_df(self, X_df, encoders=None):
        '''Transform an input dataframe, keeping track of feature names'''

        # Get as two lists - all steps but last
        ordered_objs, ordered_base_names =\
            self._get_ordered_objs_and_names()

        # Run all of the transformations
        for obj, base_name in zip(ordered_objs, ordered_base_names):
            X_df = obj.transform_df(X_df, base_name=base_name,
                                    encoders=encoders)

        return X_df

    def transform_feat_names(self, X_df, encoders=None):
        '''Like transform df, but just transform feat names.'''

        # Get as two lists - all steps but last
        ordered_objs, ordered_base_names =\
            self._get_ordered_objs_and_names()

        feat_names = list(X_df)
        for obj, base_name in zip(ordered_objs, ordered_base_names):
            feat_names = obj._proc_new_names(feat_names, base_name=base_name,
                                             encoders=encoders)

        return feat_names

    def inverse_transform_FIs(self, fis):
        '''fis should be a pandas Series as indexed by feature name'''

        # Get as two lists - all steps but last
        ordered_objs, _ =\
            self._get_ordered_objs_and_names()

        # Process each object in reverse order
        for obj in ordered_objs[::-1]:

            # Process only if object has method
            if hasattr(obj, 'inverse_transform_FIs'):
                fis = obj.inverse_transform_FIs(fis)

        return fis

    @if_delegate_has_method(delegate='_final_estimator')
    def predict(self, X, **predict_params):

        # Transform X
        Xt = self.transform(X)

        # Then return final pipeline piece predicting
        # on the transformed data
        return self.steps[-1][-1].predict(Xt, **predict_params)

    @if_delegate_has_method(delegate='_final_estimator')
    def fit_predict(self, X, y=None, **fit_params):
        raise RuntimeError('Not Implemented')

    @if_delegate_has_method(delegate='_final_estimator')
    def predict_proba(self, X):

        # Transform X and predict
        Xt = self.transform(X)
        return self.steps[-1][-1].predict_proba(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def decision_function(self, X):

        # Transform X and predict
        Xt = self.transform(X)
        return self.steps[-1][-1].decision_function(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def score_samples(self, X):

        # Transform X and score samples
        Xt = self.transform(X)
        return self.steps[-1][-1].score_samples(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def predict_log_proba(self, X):

        # Transform X and predict
        Xt = self.transform(X)
        return self.steps[-1][-1].predict_log_proba(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def score(self, X, y=None, sample_weight=None):

        # Transform X
        Xt = self.transform(X)

        # Cast y from dataframe or series if needed
        if isinstance(y, (pd.DataFrame, pd.Series)):
            y = np.array(y)

        # Rest of function
        score_params = {}
        if sample_weight is not None:
            score_params['sample_weight'] = sample_weight
        return self.steps[-1][-1].score(Xt, y, **score_params)



from ...pipeline.constructors import Constructor


class MVTransformerConstructor(Constructor):

    name = 'transformers_mv'

    def _process(self, params):

        from .options import get_mv_transformer_and_params
        from .transformer import BPtTransformerMV

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(get_mv_transformer_and_params,
                                      params)

        return self._make_col_version(objs, obj_params,
                                      params, Wrapper=BPtTransformerMV)




import pandas as pd
from BPt import Dataset


def load_boston():
    '''Loads the boston dataset using the sklearn helper function
    https://scikit-learn.org/stable/modules/generated/sklearn.datasets.load_boston.html#sklearn.datasets.load_boston
    and returns it as a BPt Dataset.'''

    from sklearn.datasets import load_boston

    # Load as dataframe
    boston_data = load_boston()
    df = pd.DataFrame(boston_data['data'],
                      columns=boston_data['feature_names'])
    df['PRICE'] = boston_data['target']

    # Convert to dataset
    dataset = Dataset(df)
    dataset = dataset.set_role('PRICE', 'target')

    # Set test split
    dataset = dataset.set_test_split(.2, random_state=3)

    return dataset




from sklearn.feature_selection import (f_regression, f_classif)
from .Params import Scalar

P = {}

# Feat Selectors
P['base univar fs regression'] = {'score_func': f_regression,
                                  'percentile': 50}

P['univar fs regression dist'] = {'score_func': f_regression,
                                  'percentile':
                                  Scalar(init=50, lower=1, upper=99)}

P['univar fs r keep more'] = {'score_func': f_regression,
                              'percentile':
                              Scalar(init=75, lower=50, upper=99)}

P['univar fs r keep less'] = {'score_func': f_regression,
                              'percentile':
                              Scalar(init=25, lower=1, upper=50)}


P['base univar fs classifier'] = {'score_func': f_classif,
                                  'percentile': 50}

P['univar fs classifier dist'] = {'score_func': f_classif,
                                  'percentile':
                                  Scalar(init=50, lower=1, upper=99)}

P['univar fs c keep more'] = {'score_func': f_classif,
                              'percentile':
                              Scalar(init=75, lower=50, upper=99)}

P['univar fs c keep less'] = {'score_func': f_classif,
                              'percentile':
                              Scalar(init=25, lower=1, upper=50)}


P['base rfe'] = {'n_features_to_select': None}

P['rfe num feats dist'] = {'n_features_to_select':
                           Scalar(init=.5, lower=.1, upper=.99)}

P['random'] = {'mask': 'sets as random features'}
P['searchable'] = {'mask': 'sets as hyperparameters'}



from ..compare import Option, Compare, Options, CompareDict


def test_option1():

    o1 = Option(value='10', name='10', key='something')
    o2 = Option(value='20', name='10', key='something')
    assert o1 == o2



from ..BPtTransformer import BPtTransformer
from sklearn.preprocessing import OneHotEncoder
import numpy as np
from BPt import Dataset
import pandas as pd


def test_all_bpt_transformer():

    X = np.array([[1, 1, 1],
                  [2, 2, 2],
                  [3, 3, 3]])
    y = np.zeros(3)

    tr = BPtTransformer(OneHotEncoder(sparse=False), inds=Ellipsis)
    tr.fit(X, y)

    assert tr.inds_ is Ellipsis
    assert len(tr.rest_inds_) == 0

    X_trans = tr.transform(X)
    assert X_trans.shape == (3, 9)
    df = pd.DataFrame(X)

    X_trans_df = tr.transform_df(df)
    assert X_trans_df.shape == (3, 9)
    for i in range(3):
        for j in range(1, 4):
            assert str(i) + '=' + str(j) in X_trans_df


def test_all_bpt_transformer_drop():

    X = np.array([[1, 1, 1],
                  [2, 2, 2],
                  [3, 3, 3]])
    y = np.zeros(3)

    tr = BPtTransformer(OneHotEncoder(sparse=False, drop='first'), 
                        inds=Ellipsis)
    tr.fit(X, y)

    assert tr.inds_ is Ellipsis
    assert len(tr.rest_inds_) == 0

    X_trans = tr.transform(X)
    assert X_trans.shape == (3, 6)
    df = pd.DataFrame(X)

    X_trans_df = tr.transform_df(df)
    assert X_trans_df.shape == (3, 6)
    for i in range(3):
        for j in range(2, 4):
            assert str(i) + '=' + str(j) in X_trans_df


def test_all_bpt_transformer_dataset():

    data = Dataset()
    data['1'] = [1, 2, 3]
    data['2'] = [4, 5, 6]
    data = data.ordinalize('all')

    assert len(data.encoders) == 2

    tr = BPtTransformer(OneHotEncoder(sparse=False), inds=Ellipsis)
    tr.fit(np.array(data), np.ones(3))

    assert tr.inds_ is Ellipsis
    assert len(tr.rest_inds_) == 0

    X_trans = tr.transform(np.array(data))
    assert X_trans.shape == (3, 6)

    # First without encoders
    X_trans_df1 = tr.transform_df(data)
    assert X_trans_df1.shape == (3, 6)
    assert '1=0' in X_trans_df1
    assert '2=2' in X_trans_df1
    assert '1=3' not in X_trans_df1
    assert '2=5' not in X_trans_df1
    assert '2=6' not in X_trans_df1

    X_trans_df2 = tr.transform_df(data, encoders=data.encoders)
    X_trans_df2.shape == (3, 6)
    assert '1=0' not in X_trans_df2
    assert '2=2' not in X_trans_df2
    assert '1=1' in X_trans_df2
    assert '1=3' in X_trans_df2
    assert '2=5' in X_trans_df2
    assert '2=6' in X_trans_df2

    assert list(X_trans_df1) != list(X_trans_df2)
    assert np.array_equal(np.array(X_trans_df1), np.array(X_trans_df2))



from joblib import Parallel, delayed, effective_n_jobs, hash as joblib_hash
from copy import deepcopy
import numpy as np
import os


class DataFile():

    def __init__(self, loc, load_func):

        self.loc = os.path.abspath(loc)
        self.load_func = load_func

    def _load(self):
        return self.load_func(self.loc)

    def load(self):
        return self._load()

    def reduce(self, reduce_func):

        # Init cache if not already
        if not hasattr(self, '_cached_reduce'):
            self._cached_reduce = {}

        # Hash the reduce func w/ joblib hash
        func_hash = joblib_hash(reduce_func)

        # If has already been hashed before
        if func_hash in self._cached_reduce:
            return self._cached_reduce[func_hash]

        # Apply the function
        reduced = reduce_func(self.load())

        # Cache result before returning
        self._cached_reduce[func_hash] = reduced

        return reduced

    def __lt__(self, other):
        return self.loc < other.loc

    def __eq__(self, other):
        return self.loc == other.loc

    def __hash__(self):
        return hash(self.loc)

    def __deepcopy__(self, memo):
        return DataFile(deepcopy(self.loc, memo), self.load_func)

    def __repr__(self):
        return 'DataFile(loc=' + repr(self.loc) + ')'

    def __str__(self):
        return self.__repr__()


def mp_single_load(files, reduce_func):

    # Create proxy to fill with values
    proxy = np.zeros(shape=(len(files)))

    # Get reduced from each file
    for f in range(len(files)):
        proxy[f] = files[f].reduce(reduce_func)

    return proxy


def load_data_file_proxy(values, reduce_func, file_mapping, n_jobs=1):

    # Replace n_jobs w/ effective n_jobs
    n_jobs = effective_n_jobs(n_jobs)

    # Can at most be number of files
    n_jobs = min([n_jobs, len(values)])

    # Create proxy to fill in
    proxy = values.copy()

    # Generate splits based on n_jobs
    splits = np.array_split(np.array(values), n_jobs)

    # Nested func for multi-proc, to vectorize
    def change_to_map(x):
        return file_mapping[x]
    v_func = np.vectorize(change_to_map)

    # Apply v_func to each split
    file_splits = [v_func(split) for split in splits]

    # Load w/ joblib Parallel
    output = Parallel(n_jobs=n_jobs,
                      backend="threading")(delayed(mp_single_load)(
                       files=files, reduce_func=reduce_func)
                      for files in file_splits)

    # Fill proxy with the concatenated output
    proxy[:] = np.concatenate(output)

    return proxy





import numpy as np
from ..FeatSelectors import FeatureSelector


def basic_test():

    mask = np.ones(10)
    mask[0] = .4
    X = np.random.random((20, 10))

    fs = FeatureSelector(mask=mask)

    X_trans = fs.fit_transform(X)

    assert X_trans.shape == (20, 9)
    assert np.array_equal(X_trans[:, 0], X[:, 1])



from sklearn.impute import SimpleImputer
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer

from ..helpers import get_obj_and_params, all_from_objects


IMPUTERS = {
    'mean': (SimpleImputer, ['mean imp']),
    'median': (SimpleImputer, ['median imp']),
    'most frequent': (SimpleImputer, ['most freq imp']),
    'constant': (SimpleImputer, ['constant imp']),
    'iterative': (IterativeImputer, ['iterative imp']),
}


def get_imputer_and_params(imputer_str, extra_params, params, **kwargs):

    imputer, extra_imputer_params, imputer_params =\
        get_obj_and_params(imputer_str, IMPUTERS, extra_params, params)

    return imputer(**extra_imputer_params), imputer_params


all_obj_keys = all_from_objects(IMPUTERS)
